from .cache import *  # will also import some globals like `britive`


def test_list():
    response = britive.system.permissions.list()
    assert isinstance(response, list)
    assert len(response) > 0
    assert isinstance(response[0], dict)
    assert 'id' in response[0].keys()
    assert 'name' in response[0].keys()


def test_create(cached_system_level_permission):
    assert isinstance(cached_system_level_permission, dict)
    assert 'id' in cached_system_level_permission.keys()
    assert 'name' in cached_system_level_permission.keys()
    assert cached_system_level_permission['name'].startswith('python-sdk')


def test_get_id(cached_system_level_permission):
    response = britive.system.permissions.get(
        permission_identifier=cached_system_level_permission['id'],
        identifier_type='id'
    )
    assert 'id' in response.keys()
    assert 'name' in response.keys()
    assert response['name'].startswith('python-sdk')


def test_get_name(cached_system_level_permission):
    response = britive.system.permissions.get(permission_identifier=cached_system_level_permission['name'])
    assert 'id' in response.keys()
    assert 'name' in response.keys()
    assert response['name'].startswith('python-sdk')


def test_update_id(cached_system_level_permission):
    permission = britive.system.permissions.build(
        name=cached_system_level_permission['name'],
        consumer='apps',
        actions=['apps.app.view', 'apps.app.manage']
    )
    response = britive.system.permissions.update(
        permission_identifier=cached_system_level_permission['id'],
        permission=permission,
        identifier_type='id'
    )

    assert response is None

    response = britive.system.permissions.get(
        permission_identifier=cached_system_level_permission['id'],
        identifier_type='id'
    )
    assert 'id' in response.keys()
    assert 'name' in response.keys()
    assert response['name'].startswith('python-sdk')
    assert len(response['actions']) == 2


def test_update_name(cached_system_level_permission):
    permission = britive.system.permissions.build(
        name=cached_system_level_permission['name'],
        consumer='apps',
        actions=['apps.app.view', 'apps.app.manage', 'apps.app.list']
    )
    response = britive.system.permissions.update(
        permission_identifier=cached_system_level_permission['name'],
        permission=permission,
        identifier_type='name'
    )

    assert response is None

    response = britive.system.permissions.get(
        permission_identifier=cached_system_level_permission['name'],
        identifier_type='name'
    )
    assert 'id' in response.keys()
    assert 'name' in response.keys()
    assert response['name'].startswith('python-sdk')
    assert len(response['actions']) == 3


def test_delete(cached_system_level_permission):
    try:
        assert britive.system.permissions.delete(
            permission_identifier=cached_system_level_permission['id'],
            identifier_type='id'
        ) is None

        with pytest.raises(exceptions.NotFound):
            britive.system.permissions.get(cached_system_level_permission['id'])
    finally:
        cleanup('permission-system-level')





class TaskServices:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/tasks/services'

    def get(self, application_id: str, service_name: str = 'environmentScanner') -> dict:
        """
        Retrieve details of the task scheduler service associated with `application_id`.

        A task service is automatically created when a new application is created. A task service cannot be created
        independent of an application.

        :param application_id: The ID of the application.
        :param service_name: The name of the task scheduler service. Currently only `environmentScanner` is supported
            but this parameter exists in case future application releases introduce other service types.
        :return: Details of the task scheduler service.
        """

        params = {
            'name': service_name,
            'appId': application_id
        }
        return self.britive.get(f'{self.base_url}', params=params)

    def enable(self, task_service_id: str) -> dict:
        """
        Enable a task service.

        :param task_service_id: The ID of the task service.
        :return: Details of the newly enabled task service.
        """

        return self.britive.post(f'{self.base_url}/{task_service_id}/enabled-statuses')

    def disable(self, task_service_id: str) -> dict:
        """
        Disable a task service.

        :param task_service_id: The ID of the task service.
        :return: Details of the newly disabled task service.
        """

        return self.britive.post(f'{self.base_url}/{task_service_id}/disabled-statuses')



class IdentityAttributes:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/users/attributes'

    def list(self):
        """
        Return a list of identity attributes.

        :return: List of identity attributes.
        """

        return self.britive.get(self.base_url)

    def create(self, name: str, description: str, data_type: str, multi_valued: bool) -> dict:
        """
        Create a new identity attribute.

        :param name: The name of the identity attribute.
        :param description: The description of the identity attribute.
        :param data_type: The data type of the identity attribute. Valid values are
            `String`, `Number`, 'Boolean`, 'Date`.
        :param multi_valued: Whether the attribute should be considered multi-valued.
        :return: Details of the newly created identity attribute.
        """

        if data_type not in ['String', 'Number', 'Boolean', 'Date']:
            raise ValueError(f'invalid data_type {data_type}')

        data = {
            'name': name,
            'description': description,
            'dataType': data_type,
            'multiValued': multi_valued
        }

        return self.britive.post(self.base_url, json=data)

    def delete(self, attribute_id: str) -> None:
        """
        Delete the specified identity attribute.

        :param attribute_id: The ID of the identity attribute to delete.
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/{attribute_id}')




from .helpers.custom_attributes import CustomAttributes
from .exceptions import NotFound


valid_statues = ['active', 'inactive']


class ServiceIdentities:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/users'
        self.custom_attributes = CustomAttributes(self)

    def list(self, filter_expression: str = None) -> list:
        """
        Provide an optionally filtered list of all service identities.

        :param filter_expression: filter list of users based on name, status, or role. The supported operators
             are 'eq' and 'co'. Example: 'name co "Smith"'
        :return: List of service identity records
        """

        params = {
            'type': 'ServiceIdentity',
            'page': 0,
            'size': 100
        }
        if filter_expression:
            params['filter'] = filter_expression

        return self.britive.get(self.base_url, params)

    def get(self, service_identity_id: str) -> dict:
        """
        Provide details of the given service_identity.

        :param service_identity_id: The ID  of the service identity.
        :return: Details of the specified user.
        """

        params = {
            'type': 'ServiceIdentity'
        }
        return self.britive.get(f'{self.base_url}/{service_identity_id}', params=params)

    def get_by_name(self, name: str) -> list:
        """
        Return service identities whose name field contains `name`.

        :param name: The name (or part of the name) of the service identity you wish to get
        :return: Details of the specified service identities. If no service identity is found will return an empty list.
        """

        service_identities = self.list(filter_expression=f'name co "{name}"')
        return service_identities

    def get_by_status(self, status: str) -> list:
        """
        Return a list of service identities filtered to `status`.

        :param status: valid values are `active` and `inactive`
        :return:
        """

        if status not in valid_statues:
            raise ValueError(f'status {status} not allowed.')

        return self.list(filter_expression=f'status eq "{status}"')

    def search(self, search_string: str) -> list:
        """
        Search all user fields for the given `search_string` and returns
        a list of matched service identities.

        :param search_string:
        :return: List of user records
        """

        params = {
            'type': 'ServiceIdentity',
            'page': 0,
            'size': 100,
            'searchText': search_string
        }

        return self.britive.get(self.base_url, params)

    def create(self, **kwargs) -> dict:
        """
        Create a new service identity.

        :param kwargs: Valid fields are...
            name - required
            description
            status - valid values are active, inactive - if omitted will default to active
        :return: Details of the newly created user.
        """

        required_fields = ['name']

        kwargs['type'] = 'ServiceIdentity'
        if 'status' not in kwargs.keys():
            kwargs['status'] = 'active'

        if kwargs['status'] not in valid_statues:
            raise ValueError(f'invalid status {kwargs["status"]}')

        if not all(x in kwargs.keys() for x in required_fields):
            raise ValueError('Not all required keyword arguments were provided.')

        response = self.britive.post(self.base_url, json=kwargs)
        return response

    def update(self, service_identity_id: str, **kwargs) -> dict:
        """
        Update the specified attributes of the provided service identity.

        Acceptable attributes are `name` and `description`.

        :param service_identity_id: The ID of the service identity to update
        :param kwargs: The attributes to update for the service identity
        :return: A dict containing the newly updated service identity details
        """

        if 'name' not in kwargs.keys():
            existing = self.get(service_identity_id)
            kwargs['name'] = existing['name']

        # add some required elements to the kwargs passed in by the caller
        kwargs['type'] = 'ServiceIdentity'
        kwargs['roleName'] = ''

        self.britive.patch(f'{self.base_url}/{service_identity_id}', json=kwargs)

        # return the updated service identity record
        return self.get(service_identity_id)

    def delete(self, service_identity_id: str) -> None:
        """
        Delete a service identity.

        :param service_identity_id: The ID of the service identity to delete
        :return: None
        """

        self.britive.delete(f'{self.base_url}/{service_identity_id}')

    def enable(self, service_identity_id: str = None, service_identity_ids: list = None) -> object:
        """
        Enable the given service identities.

        You can pass in both `service_identity_id` for a single user and `service_identity_ids` to enable multiple
        service identities in one call. If both `service_identity_id` and `service_identity_ids` are provided they
        will be merged together into one list.

        :param service_identity_id: The ID of the user you wish to enable.
        :param service_identity_ids: A list of user IDs that you wish to enable.
        :return: if `service_identity_ids` is set will return a list of user records, else returns a user dict
        """

        computed_identities = []
        if service_identity_ids:
            computed_identities += service_identity_ids
        if service_identity_id:
            computed_identities.append(service_identity_id)

        # de-dup
        computed_identities = list(set(computed_identities))
        response = self.britive.post(f'{self.base_url}/enabled-statuses', json=computed_identities)
        if not service_identity_ids:
            return response[0]
        return response

    def disable(self, service_identity_id: str = None, service_identity_ids: list = None) -> object:
        """
        Disable the given service identities.

        You can pass in both `service_identity_id` for a single service identity and `service_identity_ids` to disable
        multiple service identitie at in one call. If both `service_identity_id` and `service_identity_ids` are
        provided they will be merged together into one list.

        :param service_identity_id: The ID of the user you wish to disable.
        :param service_identity_ids: A list of user IDs that you wish to disable.
        :return: if `user_ids` is set will return a list of user records, else returns a user dict
        """

        computed_identities = []
        if service_identity_ids:
            computed_identities += service_identity_ids
        if service_identity_id:
            computed_identities.append(service_identity_id)

        # de-dup
        computed_identities = list(set(computed_identities))
        response = self.britive.post(f'{self.base_url}/disabled-statuses', json=computed_identities)
        if not service_identity_ids:
            return response[0]
        return response



from .cache import *  # will also import some globals like `britive`


tag_keys = [
    'userTagId',
    'status',
    'description',
    'name',
    'external',
    'userCount',
    'userTagIdentityProviders'
]


def test_create(cached_tag):
    assert isinstance(cached_tag, dict)
    assert set(tag_keys).issubset(cached_tag.keys())


def test_get(cached_tag):
    tag = britive.tags.get(cached_tag['userTagId'])
    assert isinstance(tag, dict)
    assert set(tag_keys).issubset(tag.keys())


def test_list(cached_tag):
    tags = britive.tags.list()
    assert isinstance(tags, list)
    assert isinstance(tags[0], dict)
    assert cached_tag['name'] in [t['name'] for t in tags]


def test_search(cached_tag):
    tags = britive.tags.search(cached_tag['name'])
    assert isinstance(tags, list)
    assert isinstance(tags[0], dict)
    assert cached_tag['name'] == tags[0]['name']


def test_users_for_tag_zero(cached_tag):
    users = britive.tags.users_for_tag(tag_id=cached_tag['userTagId'])
    assert isinstance(users, list)
    assert len(users) == 0


def test_available_users_for_tag(cached_tag):
    users = britive.tags.available_users_for_tag(tag_id=cached_tag['userTagId'])
    assert isinstance(users, list)
    assert len(users) > 0
    assert isinstance(users[0], dict)


def test_add_user(cached_tag, cached_user):
    user_added = britive.tags.add_user(tag_id=cached_tag['userTagId'], user_id=cached_user['userId'])
    assert isinstance(user_added, dict)


def test_users_for_tag_one(cached_tag):
    users = britive.tags.users_for_tag(tag_id=cached_tag['userTagId'])
    assert isinstance(users, list)
    assert len(users) == 1
    assert isinstance(users[0], dict)


def test_remove_user(cached_tag, cached_user):
    response = britive.tags.remove_user(tag_id=cached_tag['userTagId'], user_id=cached_user['userId'])
    assert response is None


def test_enable(cached_tag):
    response = britive.tags.enable(cached_tag['userTagId'])
    assert isinstance(response, dict)
    assert response['userTagId'] == cached_tag['userTagId']
    assert response['status'] == 'Active'


def test_disable(cached_tag):
    response = britive.tags.disable(cached_tag['userTagId'])
    assert isinstance(response, dict)
    assert response['userTagId'] == cached_tag['userTagId']
    assert response['status'] == 'Inactive'


def test_update(cached_tag):
    r = str(random.randint(0, 1000000))
    tag = britive.tags.update(cached_tag['userTagId'], name=f'testpythonapteiwrappertag-{r}')
    assert isinstance(tag, dict)
    assert set(tag_keys).issubset(tag.keys())
    assert tag['name'] == f'testpythonapteiwrappertag-{r}'
    # set it back for downstream processes
    britive.tags.update(cached_tag['userTagId'], name=cached_tag['name'])


def test_membership_rules_list(cached_tag):
    response = britive.tags.membership_rules.list(tag_id=cached_tag['userTagId'])
    assert len(response) == 0


def test_membership_rules_create(cached_tag, cached_user):
    rules = [
        britive.tags.membership_rules.build(attribute_id_or_name='Email', operator='is', value=cached_user['email'])
    ]

    response = britive.tags.membership_rules.create(tag_id=cached_tag['userTagId'], rules=rules)
    assert len(response) == 1

    response = britive.tags.membership_rules.list(tag_id=cached_tag['userTagId'])
    assert len(response) == 1


def test_membership_rules_update(cached_tag, cached_user):
    rules = [
        britive.tags.membership_rules.build(
            attribute_id_or_name='Email',
            operator='is',
            value=cached_user['email']
        ),
        britive.tags.membership_rules.build(
            attribute_id_or_name='Username',
            operator='is',
            value=cached_user['username']
        )
    ]

    response = britive.tags.membership_rules.update(tag_id=cached_tag['userTagId'], rules=rules)
    assert response is None

    response = britive.tags.membership_rules.list(tag_id=cached_tag['userTagId'])
    assert len(response) == 2


def test_membership_rules_matched_users(cached_tag, cached_user):
    response = britive.tags.membership_rules.matched_users(tag_id=cached_tag['userTagId'])
    assert len(response) == 1
    assert response[0]['email'] == cached_user['email']


def test_membership_rules_delete(cached_tag):
    response = britive.tags.membership_rules.delete(tag_id=cached_tag['userTagId'])
    assert response is None

    response = britive.tags.membership_rules.list(tag_id=cached_tag['userTagId'])
    assert len(response) == 0


def test_minimized_user_details(cached_tag):
    details = britive.tags.minimized_tag_details(tag_id=cached_tag['userTagId'])
    assert isinstance(details, list)
    assert len(details) == 1
    details = britive.tags.minimized_tag_details(tag_ids=[cached_tag['userTagId']])
    assert isinstance(details, list)
    assert len(details) == 1



# note that the order of deletes could matter here so we will delete resources
# in the opposite order of creation.

from distutils.command.clean import clean
from typing import final
from .cache import *  # will also import some globals like `britive`
from britive import exceptions


# 130-profiles
def test_profile_delete(cached_profile):
    try:
        response = britive.profiles.delete(
            application_id=cached_profile['appContainerId'],
            profile_id=cached_profile['papId']
        )

        profiles = britive.profiles.list(application_id=cached_profile['appContainerId'])

        assert response is None
        assert cached_profile['papId'] not in [p['papId'] for p in profiles]
    finally:
        cleanup('profile')
        cleanup('profile-approval-policy')
        cleanup('dynamic-session-attribute')
        cleanup('static-session-attribute')


# 070-environments
def test_environment_delete(cached_application, cached_environment):
    try:
        response = britive.environments.delete(
            application_id=cached_application['appContainerId'],
            environment_id=cached_environment['id']
        )
        assert response is None
    finally:
        cleanup('environment')
        cleanup('scan')


# 060-environment_groups
def test_environment_group_delete(cached_application, cached_environment_group):
    try:
        groups = britive.environment_groups.list(application_id=cached_application['appContainerId'])
        num_root_groups = 0
        for group in groups:
            if group['parentId'] != '':  # cannot delete root groups - error A-0003 is thrown when attempting
                response = britive.environment_groups.delete(
                    application_id=cached_application['appContainerId'],
                    environment_group_id=cached_environment_group['id']
                )
                assert response is None
            else:
                num_root_groups += 1
        groups = britive.environment_groups.list(application_id=cached_application['appContainerId'])
        assert isinstance(groups, list)
        assert len(groups) == num_root_groups  # the root group will remain
        assert isinstance(groups[0], dict)
    finally:
        cleanup('environment-group')


# 050-applications
def test_application_delete(cached_application):
    try:
        response = britive.applications.delete(application_id=cached_application['appContainerId'])
        assert response is None
    finally:
        cleanup('application')
        cleanup('catalog')
        cleanup('account')
        cleanup('group')
        cleanup('permission')
        cleanup('task-service')


# 040-service_identity_tokens
def test_service_identity_tokens_delete(cached_service_identity):
    try:
        britive.service_identities.delete(cached_service_identity['userId'])
        with pytest.raises(exceptions.InvalidRequest):
            britive.service_identity_tokens.get(cached_service_identity['userId'])
    finally:
        cleanup('service-identity-token')
        cleanup('service-identity-token-updated')


# 030-service_identities
def test_service_identities_delete(cached_service_identity):
    try:
        response = britive.service_identities.delete(service_identity_id=cached_service_identity['userId'])
        assert response is None

        with pytest.raises(exceptions.NotFound):
            britive.service_identities.get_by_name(name=cached_service_identity['name'])
    except exceptions.NotFound:
        pass
    finally:
        cleanup('service-identity')


# 020-tags
def test_tags_delete(cached_tag):
    try:
        response = britive.tags.delete(cached_tag['userTagId'])
        assert response is None
    finally:
        cleanup('tag')


# 010-users
def test_user_delete(cached_user):
    try:
        response = britive.users.delete(user_id=cached_user['userId'])
        assert response is None
        users = britive.users.get_by_name(name=cached_user['lastName'])
        assert isinstance(users, list)
        assert len(users) == 0
    finally:
        cleanup('user')


# 005-identity_attributes
def test_identity_attribute_delete(cached_identity_attribute):
    try:
        response = britive.identity_attributes.delete(attribute_id=cached_identity_attribute['id'])
        assert response is None
    finally:
        cleanup('identity-attribute')


def test_folder_delete(cached_folder, cached_vault):
    try:
        response = britive.secrets_manager.folders.delete(
            path=cached_folder['path'], vault_id=cached_vault['id'])
        assert response is None
    finally:
        cleanup('folder')


def test_secret_delete(cached_secret, cached_vault):
    try:
        response = britive.secrets_manager.secrets.delete(
            path=cached_secret['path'],
            vault_id=cached_vault['id']
            )
        assert response is None
    finally:
        cleanup('secret')


def test_static_secret_templates_delete(cached_static_secret_template):
    try:
        response = britive.secrets_manager.static_secret_templates.delete(
            cached_static_secret_template['id'])
        assert response is None
    finally:
        cleanup('static-secret-templates')


def test_passwordPolicies_delete(cached_PasswordPolicies):
    try:
        response = britive.secrets_manager.password_policies.delete(cached_PasswordPolicies['id'])
        assert response is None
    finally:
        cleanup('password-policies')


def test_pinPolicy_delete(cached_PinPolicies):
    try:
        response = britive.secrets_manager.password_policies.delete(cached_PinPolicies['id'])
        assert response is None
    finally:
        cleanup('pin-policies')


def test_policy_delete(cached_policy):
    try:
        response = britive.secrets_manager.policies.delete(cached_policy['id'])
        assert response is None
    finally:
        cleanup('policy')


def test_vault_delete(cached_vault):
    try:
        response = britive.secrets_manager.vaults.delete(cached_vault['id'])
        assert response is None
    finally:
        cleanup('vault')


def test_notification_medium_delete(cached_notification_medium):
    try:
        response = britive.notification_mediums.delete(cached_notification_medium['id'])
        assert response is None
    finally:
        cleanup('notification-medium')
        


import json
from typing import Union


class SystemPolicies:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/v1/policy-admin/policies'

    @staticmethod
    def _validate_identifier_type(identifier_type):
        if identifier_type not in ['id', 'name']:
            raise ValueError(f'identifier_type of {identifier_type} is invalid. Only `name` and `id` are allowed.')

    def list(self, filter_expression: str = '') -> list:
        """
        List system level policies (not including policies for secrets manager or profiles).

        :param filter_expression: Filter based on `name`. Valid operators are `eq`, `sw`, and
            `co`. Example: name co policy
        :returns: List of policies.
        """

        params = {}
        if filter_expression:
            params['filter'] = filter_expression
        return self.britive.get(self.base_url, params=params)

    def get(self, policy_identifier: str, identifier_type: str = 'name', verbose: bool = False,
            condition_as_dict: bool = False) -> dict:
        """
        Get details of the specified system policy.

        :param policy_identifier: The ID or name of the policy.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers will be returned. Either all identifiers must be names or all identifiers must be IDs.
        :param verbose: Whether to return a more compact response (the default) or a more verbose response.
        :param condition_as_dict: Prior to version 2.22.0 a policy condition block was always returned as stringified
            json. As of 2.22.0 the SDK now supports returning the condition block of a policy as either stringified json
            or a raw python dictionary. The Britive backend will also return the condition block in either format,
            depending on a query parameter value. Setting this value to `True` will result in the condition block being
            returned as a python dictionary. The default of `False` is to support backwards compatibility.
        :return: Details of the specified policy.
        """

        self._validate_identifier_type(identifier_type)

        params = {
            'compactResponse': not verbose,
            'conditionJson': condition_as_dict
        }

        return self.britive.get(f'{self.base_url}/{policy_identifier}', params=params)

    def create(self, policy: dict) -> dict:
        """
        Create a system level policy.

        :param policy: The policy to create. Use `policies.build` to assist in constructing a proper policy document.
        :returns: Details of the newly created policy.
        """

        return self.britive.post(self.base_url, json=policy)
    
    def update(self, policy_identifier: str, policy: dict, identifier_type: str = 'name') -> None:
        """
        Update a system level policy.

        :param policy_identifier: The ID or name of the policy to update.
        :param policy: The policy to update. Use `policies.build` to assist in constructing a proper policy document.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers will be returned. Either all identifiers must be names or all identifiers must be IDs.
        :returns: None.
        """

        self._validate_identifier_type(identifier_type)
        return self.britive.patch(f'{self.base_url}/{policy_identifier}', json=policy)

    def delete(self, policy_identifier: str, identifier_type: str = 'name') -> None:
        """
        Delete a system level policy.

        :param policy_identifier: The ID or name of the policy to delete.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers will be returned. Either all identifiers must be names or all identifiers must be IDs.
        :returns: None.
        """

        self._validate_identifier_type(identifier_type)
        return self.britive.delete(f'{self.base_url}/{policy_identifier}')

    def disable(self, policy_identifier: str, identifier_type: str = 'name') -> None:
        """
        Disable a system level policy.

        :param policy_identifier: The ID of the policy to disable.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers will be returned. Either all identifiers must be names or all identifiers must be IDs.
        :returns: None.
        """

        self._validate_identifier_type(identifier_type)
        return self.britive.patch(f'{self.base_url}/{policy_identifier}', json={'isActive': False})

    def enable(self, policy_identifier: str, identifier_type: str = 'name') -> None:
        """
        Enable a system level policy.

        :param policy_identifier: The ID of the policy to enable.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers will be returned. Either all identifiers must be names or all identifiers must be IDs.
        :returns: None.
        """

        self._validate_identifier_type(identifier_type)
        return self.britive.patch(f'{self.base_url}/{policy_identifier}', json={'isActive': True})

    def evaluate(self, statements: list) -> dict:
        """
        Evaluate the calling identities access for the given set of statements.

        :param: statements: List of statements in the following format. Use `build_evaluate_statement` to help
            construct the statement list.
            [
                {
                    'action': '<action>',
                    'resource': '<resource>',
                    'consumer': '<consumer>'
                },
            ]
        :returns: Dictionary containing each statement mapped to Allow or Deny.
        """

        return self.britive.post(f'{self.britive.base_url}/v1/policy-admin/batchevaluate', json=statements)

    @staticmethod
    def build_evaluate_statement(consumer: str, action: str, resource: str = '*') -> dict:
        """
        Builds a statement which can be evaluated with `evaluate`.

        :param consumer: The consumer for the statement.
        :param action: The action for the statement.
        :param resource: Optional resource for the statement. Defaults to `*`.
        :returns: The statement.
        """
        return {
            'action': action,
            'resource': resource,
            'consumer': consumer
        }

    @staticmethod
    def _time_of_access_type_from_deprecated_fields(from_time: str, to_time: str):
        if len(from_time) == 8 and len(to_time) == 8:
            return 'days'
        if len(from_time) == 19 and len(to_time) == 19:
            return 'date'
        raise ValueError('from_time and to_time must both be valid times or datetimes. Cannot mix.')

    @staticmethod
    def build(name: str, description: str = '', draft: bool = False, active: bool = True,
              read_only: bool = False, users: list = None, tags: list = None, tokens: list = None,
              service_identities: list = None, permissions: list = None, roles: list = None, ips: list = None,
              from_time: str = None, to_time: str = None, date_schedule: dict = None, days_schedule: dict = None,
              approval_notification_medium: Union[str, list] = None, time_to_approve: int = 5,
              access_validity_time: int = 120, approver_users: list = None, approver_tags: list = None,
              access_type: str = 'Allow', identifier_type: str = 'name', condition_as_dict: bool = False) -> dict:
        """
        Build a policy document given the provided inputs.

        :param name: The name of the policy.
        :param description: An optional description of the policy.
        :param draft: Indicates if the policy is a draft. Defaults to `False`.
        :param active: Indicates if the policy is active. Defaults to `True`.
        :param read_only: Indicates if the policy is a read only. Defaults to `False`.
        :param users: Optional list of user names or ids to which this policy applies.
        :param tags: Optional list of tag names or ids to which this policy applies.
        :param tokens: Optional list of token names or ids to which this policy applies.
        :param service_identities: Optional list of service identity names or ids to which this policy applies.
        :param permissions: Optional list of permission names or ids this policy grants. Provide either this parameter
            or `roles`.
        :param roles: Optional list of role names or ids to which this policy applies. Provider either this parameter
            or `permissions`.
        :param ips: Optional list of IP addresses for which this policy applies. Provide in CIDR notation
            or dotted decimal format for individual (/32) IP addresses.
        :param from_time: The start date/time of when the policy is in effect. If a date is provided
            (`YYYY-MM-DD HH:MM:SS`) this will represent the start date/time of 1 contiguous time range. If just a
            time is provided (`HH:MM:SS`) this will represent the daily recurring start time. If this parameter is
            provided then `to_time` must also be provided. This parameter is deprecated as of v2.19.0. The presence of
            `date_schedule` and/or `days_schedule` will override this field.
        :param to_time: The end date/time of when the policy is in effect. If a date is provided
            (`YYYY-MM-DD HH:MM:SS`) this will represent the end date/time of 1 contiguous time range. If just a
            time is provided (`HH:MM:SS`) this will represent the daily recurring end time. If this parameter is
            provided then `from_time` must also be provided. This parameter is deprecated as of v2.19.0. The presence of
            `date_schedule` and/or `days_schedule` will override this field.
        :param date_schedule: A dict in the format
            {
                'fromDate': '2022-10-29 10:30:00',
                'toDate': '2022-11-05 18:30:00',
                'timezone': 'UTC'
            }
            Timezone formats can be found in the TZ Identifier column of the following page.
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
        :param days_schedule: A dict in the format
            {
                'fromTime': '10:30:00',
                'toTime': '18:30:00',
                'timezone': 'UTC',
                'days': ['MONDAY','TUESDAY','WEDNESDAY','THURSDAY','FRIDAY','SATURDAY','SUNDAY']
            }
            Timezone formats can be found in the TZ Identifier column of the following page.
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
        :param approval_notification_medium: Optional notification medium name to which approval requests will be
            delivered. Can also specify a list of notification medium names. Specifying this parameter indicates the
            desire to enable approvals for this policy.
        :param time_to_approve: Optional number of minutes to wait for an approval before denying the action. Defaults
            to 5 minutes.
        :param access_validity_time: Optional number of minutes the access is valid after approval. Defaults to 120
            minutes.
        :param approver_users: Optional list of user names or ids who are to be considered approvers.
            If `approval_notification_medium` is set then either `approver_users` or `approver_tags` is required.
        :param approver_tags: Optional list of tag names who are considered approvers.
            If `approval_notification_medium` is set then either `approver_users` or `approver_tags` is required.
        :param access_type: The type of access this policy provides. Valid values are `Allow` and `Deny`. Defaults
            to `Allow`.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers are being provided to the other parameters. Either all identifiers must be names or all
            identifiers must be IDs.
        :param condition_as_dict: Prior to version 2.22.0 the only acceptable format for the condition block of
            a policy was as a stringifed json object. As of 2.22.0 the condition block can also be built as a raw
            python dictionary. This parameter will default to `False` to support backwards compatibility. Setting to
            `True` will result in the policy condition being returned/built as a python dictionary.
        :return: A dict which can be provided as a policy to `create` and `update`.
        """

        condition = {}

        # handle ip address logic
        if ips:
            condition['ipAddress'] = ','.join(ips)

        # handle from_time and to_time logic
        # this logic is deprecated (and updated) in v2.19.0 and will be retired in v3.x.x
        if from_time and not to_time:
            raise ValueError('if from_time is provided then to_time must also be provided.')
        if to_time and not from_time:
            raise ValueError('if to_time is provided then from_time must also be provided.')
        if from_time and to_time:
            # we need to determine if we are dealing with dates or times
            # as that will inform if we set days_schedule or date_schedule
            time_of_access_type = SystemPolicies._time_of_access_type_from_deprecated_fields(from_time, to_time)
            if time_of_access_type == 'days' and not days_schedule:  # dealing with times of day
                days_schedule = {
                    'fromTime': from_time,
                    'toTime': to_time,
                    'timezone': 'UTC',
                    'days': [
                        'MONDAY',
                        'TUESDAY',
                        'WEDNESDAY',
                        'THURSDAY',
                        'FRIDAY',
                        'SATURDAY',
                        'SUNDAY'
                    ]
                }
            if time_of_access_type == 'date' and not date_schedule:  # dealing with date range
                date_schedule = {
                    'fromDate': from_time,
                    'toDate': to_time,
                    'timezone': 'UTC'
                }

        if date_schedule or days_schedule:
            condition['timeOfAccess'] = {
                'dateSchedule': date_schedule,
                'daysSchedule': days_schedule
            }

        # handle approval logic
        if approval_notification_medium:
            if not approver_users and not approver_tags:
                raise ValueError('when approval is required either approver_tags or approver_users or both '
                                 'must be provided')
            approval_condition = {
                'notificationMedium': approval_notification_medium,
                'timeToApprove': time_to_approve,
                'validFor': access_validity_time,
                'isValidForInDays': False,  # the SDK will only support minutes
                'approvers': {
                    'userIds': approver_users,
                    'tags': approver_tags
                }
            }

            if not approver_users:
                approval_condition['approvers'].pop('userIds')
            if not approver_tags:
                approval_condition['approvers'].pop('tags')

            condition['approval'] = approval_condition

        # put it all together
        policy = {
            'name': name,
            'description': description,
            'isActive': active,
            'isDraft': draft,
            'isReadOnly': read_only,
            'accessType': access_type,
            'members': {
                'users': [{identifier_type: u} for u in users] if users else None,
                'tags': [{identifier_type: t} for t in tags] if tags else None,
                'serviceIdentities': [{identifier_type: s} for s in service_identities] if service_identities else None,
                'tokens': [{identifier_type: t} for t in tokens] if tokens else None
            }
        }

        if not users:
            policy['members'].pop('users')
        if not tags:
            policy['members'].pop('tags')
        if not service_identities:
            policy['members'].pop('serviceIdentities')
        if not tokens:
            policy['members'].pop('tokens')

        if permissions:
            policy['permissions'] = [{identifier_type: p} for p in permissions]
        if roles:
            policy['roles'] = [{identifier_type: r} for r in roles]
        if condition:
            policy['condition'] = condition if condition_as_dict else json.dumps(condition, default=str)

        return policy



import datetime
from .cache import *  # will also import some globals like `britive`


def test_get():
    banner = britive.settings.banner.get()
    assert isinstance(banner, dict)


def test_set_no_schedule():
    banner = britive.settings.banner.set(message='test', display_banner=True, message_type='INFO')
    assert isinstance(banner, dict)
    for key in ['status', 'messageType', 'message']:
        assert key in banner
    assert 'messageSchedule' not in banner


def test_set_with_schedule():
    today = datetime.date.today()
    start_datetime = datetime.date.today()
    banner = britive.settings.banner.set(
        message='test',
        display_banner=True,
        message_type='INFO',
        start_datetime=start_datetime,
        end_datetime=start_datetime + datetime.timedelta(days=1),
        time_zone='UTC'
    )
    assert isinstance(banner, dict)
    for key in ['status', 'messageType', 'message', 'messageSchedule']:
        assert key in banner
    for key in ['startDate', 'endDate', 'timeZone']:
        assert key in banner['messageSchedule']


def test_set_with_incorrect_schedule():
    with pytest.raises(ValueError):
        britive.settings.banner.set(
            message='test',
            display_banner=True,
            message_type='INFO',
            start_datetime=datetime.datetime(year=2024, month=1, day=1),
            time_zone='UTC'
        )

    with pytest.raises(ValueError):
        britive.settings.banner.set(
            message='test',
            display_banner=True,
            message_type='INFO',
            end_datetime=datetime.datetime(year=2024, month=1, day=1),
            time_zone='UTC'
        )

    with pytest.raises(ValueError):
        britive.settings.banner.set(
            message='test',
            display_banner=True,
            message_type='INFO',
            start_datetime=datetime.datetime(year=2024, month=1, day=1),
            end_datetime=datetime.datetime(year=2024, month=1, day=1)
        )

    with pytest.raises(ValueError):
        britive.settings.banner.set(
            message='test',
            display_banner=True,
            message_type='INFO',
            start_datetime=datetime.datetime(year=2024, month=1, day=1)
        )

    with pytest.raises(ValueError):
        britive.settings.banner.set(
            message='test',
            display_banner=True,
            message_type='INFO',
            end_datetime=datetime.datetime(year=2024, month=1, day=1)
        )

    with pytest.raises(ValueError):
        britive.settings.banner.set(
            message='test',
            display_banner=True,
            message_type='INFO',
            time_zone='UTC'
        )


def test_banner_end_user():
    banner = britive.banner()
    assert isinstance(banner, dict)
    assert 'message' in banner
    assert 'messageType' in banner


def test_banner_off():
    banner = britive.settings.banner.set(display_banner=False, message='dont care', message_type='INFO')
    assert 'status' in banner
    assert banner['status'] == 'OFF'
    banner = britive.banner()
    assert banner is None




class Accounts:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/apps'

    def list(self, application_id: str, environment_id: str = None,
             include_associations: bool = True, filter_expression: str = None) -> list:
        """
        Return details of all the accounts associated with a given application and environment.

        :param application_id: The ID of the application.
        :param environment_id: Optionally the ID of the environment. Required only for applications which have
            `catalogApplication.supportsEnvironmentScanning` set to `False`.
        :param include_associations: Include the associated permissions and groups in the response.
        :param filter_expression: Filter based on `scanStatus`. Example: `scanStatus eq Unchanged`.
        :return: List containing details of each account.
        """

        params = {
            'page': 0,
            'size': 100,
            'includeMembers': include_associations
        }

        environment_id = environment_id or self.britive.get_root_environment_group(application_id)

        if filter_expression:
            params['filter'] = filter_expression

        url = f'{self.base_url}/{application_id}/environments/{environment_id}/accounts'
        return self.britive.get(url, params=params)

    def permissions(self, account_id: str, application_id: str,
                    environment_id: str = None, filter_expression: str = None) -> list:
        """
        Return details about the permissions associated to the specified account.

        :param account_id: The ID of the account.
        :param application_id: The ID of the application.
        :param environment_id: Optionally the ID of the environment. Required only for applications which have
            `catalogApplication.supportsEnvironmentScanning` set to `False`.
        :param filter_expression: Filter based on `scanStatus`. Example: `scanStatus eq Unchanged`.
        :return: List of permissions associated with the specified account.
        """

        params = {
            'page': 0,
            'size': 100
        }

        environment_id = environment_id or self.britive.get_root_environment_group(application_id)

        if filter_expression:
            params['filter'] = filter_expression

        url = f'{self.base_url}/{application_id}/environments/{environment_id}/accounts/{account_id}/permissions'
        return self.britive.get(url, params=params)

    def groups(self, account_id: str, application_id: str,
               environment_id: str = None, filter_expression: str = None) -> list:
        """
        Return details about the groups associated to the specified account.

        :param account_id: The ID of the account.
        :param application_id: The ID of the application.
        :param environment_id: Optionally the ID of the environment. Required only for applications which have
            `catalogApplication.supportsEnvironmentScanning` set to `False`.
        :param filter_expression: Filter based on `scanStatus`. Example: `scanStatus eq Unchanged`.
        :return: List of groups associated with the specified account.
        """

        params = {
            'page': 0,
            'size': 100
        }

        environment_id = environment_id or self.britive.get_root_environment_group(application_id)

        if filter_expression:
            params['filter'] = filter_expression

        url = f'{self.base_url}/{application_id}/environments/{environment_id}/accounts/{account_id}/groups'
        return self.britive.get(url, params=params)

    def map(self, user_id: str, account_id: str, application_id: str, environment_id: str = None,
            map_user_to_account_in_all_application_environments: bool = False) -> list:
        """
        Associate a user (or users) with an account.

        :param user_id: The ID of the user you wish to map.
        :param account_id: The ID of the account.
        :param application_id: The ID of the application.
        :param environment_id:Optionally the ID of the environment. Required only for applications which have
            `catalogApplication.supportsEnvironmentScanning` set to `False`.
        :param map_user_to_account_in_all_application_environments: If set to True will map the user to the specified
            account across all environments of the application.
        :return: List of mapped users for the account.
        """

        environment_id = environment_id or self.britive.get_root_environment_group(application_id)

        url = f'{self.base_url}/{application_id}/environments/{environment_id}/accounts/{account_id}/users'

        if map_user_to_account_in_all_application_environments:
            url += f'/{user_id}'
            return self.britive.post(url, json={'saveToAllEnvs': True})
        else:
            return self.britive.post(url, json=[user_id])

    def unmap(self, user_id: str, account_id: str, application_id: str,
              environment_id: str = None, force_profile_checkins: bool = True) -> list:
        """
        Remove a user from an account.

        :param user_id: The ID of the user you wish to map.
        :param account_id: The ID of the account.
        :param application_id: The ID of the application.
        :param environment_id: Optionally the ID of the environment. Required only for applications which have
            `catalogApplication.supportsEnvironmentScanning` set to `False`.
        :param force_profile_checkins: Force any currently checked out profiles for the user to be auto checked in.
            Will only check in profiles that are associated with the account being unmapped. Other profiles the user
            has will not be impacted.
        :return: List of mapped users for the account.
        """

        environment_id = environment_id or self.britive.get_root_environment_group(application_id)

        params = {
            'forceCheckinPaps': force_profile_checkins
        }

        url = f'{self.base_url}/{application_id}/environments/{environment_id}/accounts/{account_id}/users/{user_id}'
        return self.britive.delete(url, params=params)

    def mapped_users(self, account_id: str, application_id: str, environment_id: str = None) -> list:
        """
        Return a list of users who are mapped to the specified account.

        :param account_id: The ID of the account.
        :param application_id: The ID of the application.
        :param environment_id: Optionally the ID of the environment. Required only for applications which have
            `catalogApplication.supportsEnvironmentScanning` set to `False`.
        :return: List of users mapped to the specified account.
        """

        environment_id = environment_id or self.britive.get_root_environment_group(application_id)

        url = f'{self.base_url}/{application_id}/environments/{environment_id}/accounts/{account_id}/users'
        return self.britive.get(url)

    def users_available_to_map(self, account_id: str, application_id: str, environment_id: str = None) -> list:
        """
        Return a list of users available to be mapped to the specified account.

        :param account_id: The ID of the account.
        :param application_id: The ID of the application.
        :param environment_id: Optionally the ID of the environment. Required only for applications which have
            `catalogApplication.supportsEnvironmentScanning` set to `False`.
        :return: List of users available to be mapped to the specified account.
        """

        params = {
            'page': 0,
            'size': 100,
            'query': 'available'
        }

        environment_id = environment_id or self.britive.get_root_environment_group(application_id)

        url = f'{self.base_url}/{application_id}/environments/{environment_id}/accounts/{account_id}/users'
        return self.britive.get(url, params=params)




class SecurityPolicies:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/security-policies'

    def list(self) -> list:
        """
        Retrieve the list of all security policies in the system.

        :return: List of security policies.
        """

        return self.britive.get(self.base_url)

    def get(self, security_policy_id: str) -> dict:
        """
        Retrieve the details of the specified security policy.

        :param security_policy_id: The ID of the security policy.
        :return: Details of the security policy
        """

        return self.britive.get(f'{self.base_url}/{security_policy_id}')

    def create(self, name: str, effect: str, ips: list, tokens: list, description: str = None) -> dict:
        """
        Create a new security policy.

        Security policies are applied to API tokens and are currently only able to restrict usage to a set of IP
        addresses.

        :param name: Name of the security policy.
        :param effect: Valid values are `Allow` and `Deny`.
        :param ips: List of IP addresses in CIDR format (x.x.x.x/y). Individual IP addresses are also allowed.
        :param tokens: List of API token IDs for which this security policy should be applied.
        :param description: Optional description of the security policy.
        :return: Details of the newly created security policy.
        """

        data = {
            'name': name,
            'description': description or '',
            'effect': effect,
            'conditions': [
                {
                    'values': ips,
                    'type': 'ipAddress',
                    'evaluation': 'any'
                }
            ],
            'assignedTokens': [{'id': t} for t in tokens]
        }
        return self.britive.post(self.base_url, json=data)

    def enable(self, security_policy_id: str) -> None:
        """
        Enable a security policy.

        :param security_policy_id: The ID of the security policy.
        :return: None
        """

        return self.britive.patch(f'{self.base_url}/{security_policy_id}/enabled-statuses')

    def disable(self, security_policy_id: str) -> None:
        """
        Disable a security policy.

        :param security_policy_id: The ID of the security policy.
        :return: None
        """

        return self.britive.patch(f'{self.base_url}/{security_policy_id}/disabled-statuses')

    def update(self, security_policy_id: str, name: str = None, effect: str = None, ips: list = None,
               tokens: list = None, description: str = None) -> None:
        """
        Update a security policy.

        Only the parameters provided will be updated.

        :param security_policy_id: The ID of the security policy to update.
        :param name: Name of the security policy.
        :param effect: Valid values are `Allow` and `Deny`.
        :param ips: List of IP addresses in CIDR format (x.x.x.x/y). Individual IP addresses are also allowed.
        :param tokens: List of API token IDs for which this security policy should be applied.
        :param description: Optional description of the security policy.
        :return: Details of the newly created security policy.
        """

        # this update method is super strange - it doesn't following any of the standard update logic that other
        # resource types use.

        # first lets grab the existing security policy and then we can modify it based on what parameters have been
        # passed in.

        existing_policy = self.get(security_policy_id=security_policy_id)

        # construct part of the dict assuming every parameter is present
        raw_data = {
            'id': security_policy_id,
            'name': name,
            'description': description,
            'effect': effect
        }

        # and now remove any None values
        update = {k: v for k, v in raw_data.items() if v is not None}

        # handle the more complex attributes if needed
        if ips:
            update['conditions'] = [
                {
                    'values': ips,
                    'type': 'ipAddress',
                    'evaluation': 'any'
                }
            ]
        if tokens:
            update['assignedTokens'] = [{'id': t} for t in tokens]

        # and finally overwrite old values with new values, if they exist
        data = {**existing_policy, **update}
        
        return self.britive.put(f'{self.base_url}', json=data)

    def delete(self, security_policy_id: str) -> None:
        """
        Delete a security policy.

        :param security_policy_id: The ID of the security policy to update.
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/{security_policy_id}')



from .cache import *  # will also import some globals like `britive`


def test_create(cached_api_token):
    assert isinstance(cached_api_token, dict)
    assert 'token' in cached_api_token.keys()


def test_list(cached_api_token):
    tokens = britive.api_tokens.list()
    assert isinstance(tokens, list)
    assert cached_api_token['id'] in [t['id'] for t in tokens]


def test_get(cached_api_token):
    token = britive.api_tokens.get(token_id=cached_api_token['id'])
    assert isinstance(token, dict)
    assert token['id'] == cached_api_token['id']


def test_update(cached_api_token):
    r = str(random.randint(0, 1000000))
    new_name = f'test-{r}'
    response = britive.api_tokens.update(token_id=cached_api_token['id'], name=new_name)
    assert response is None
    response = britive.api_tokens.update(token_id=cached_api_token['id'], expiration_days=10)
    assert response is None
    token = britive.api_tokens.get(token_id=cached_api_token['id'])
    assert token['tokenExpirationDays'] == 10
    assert token['name'] == new_name


def test_delete(cached_api_token):
    response = britive.api_tokens.delete(token_id=cached_api_token['id'])
    assert response is None
    cleanup('api-token')



from . import exceptions
from typing import Union


class Workload:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/workload'
        self.identity_providers = self.IdentityProviders(self)
        self.service_identities = self.ServiceIdentities(self)
        self.scim_user = self.ScimUser(self)

    class IdentityProviders:
        def __init__(self, workload):
            self.britive = workload.britive
            self.base_url = f'{workload.base_url}/identity-providers'

        def list(self, idp_type: str = None) -> list:
            """
            Return a list of all workload identity providers.

            :param idp_type: Optional filter to apply to reduce the results to a specific workload identity provider
                type. Valid values are `AWS` and `OIDC`.
            :returns: List of all workload identity providers.
            """

            params = {}
            if idp_type:
                params['type'] = idp_type
            return self.britive.get(self.base_url, params=params)

        def get(self, workload_identity_provider_id: int) -> dict:
            """
            Return details of the specified workload identity provider.

            :param workload_identity_provider_id: The ID of the workload identity provider.
            :returns: Details of the specified workload identity provider.
            """

            return self.britive.get(f'{self.base_url}/{workload_identity_provider_id}')

        def _build_attributes_map_list(self, attributes_map: dict):
            # first get list of existing custom identity attributes and build some helpers
            existing_attrs = [attr for attr in self.britive.identity_attributes.list() if not attr['builtIn']]
            existing_attr_ids = [attr['id'] for attr in existing_attrs]
            attrs_by_name = {attr['name']: attr['id'] for attr in existing_attrs}

            # for each attributeMap key/value provided ensure we convert to ID and build the list
            attrs_list = []
            for idp_attr, custom_identity_attribute in attributes_map.items():
                if custom_identity_attribute not in existing_attr_ids:
                    custom_identity_attribute = attrs_by_name.get(custom_identity_attribute, None)
                if not custom_identity_attribute:
                    raise ValueError(f'custom identity attribute name {custom_identity_attribute} not found.')
                attrs_list.append(
                    {
                        'idpAttr': idp_attr,
                        'userAttr': custom_identity_attribute
                    }
                )
            return attrs_list

        def create(self, **kwargs) -> dict:
            """
            Create a workload identity provider.

            This method accepts a list of kwargs which match the required and optional fields for creating a
            workload identity provider. It mostly exists to support the creation of new workload identity provider
            types before this SDK can be updated to support the new type natively.

            Generally, the caller should opt to use the `create_aws` or `create_oidc` methods instead of calling
            this method directly.

            The field `attributesMap` is in format `{'idp_attr_value': 'custom_identity_attribute_name_or_id', ...}`.

            :param kwargs: A list of keyword arguments which will be provided directly to the API backend without
                any further inspection. Valid fields are `name`, `description`, `idpType`, `attributesMap`, and
                `validationWindow`. For the AWS provider an additional field `maxDuration` is valid. For the OIDC
                provider additional fields `issuerUrl` and `allowedAudiences` are valid.
            :returns: Details of the newly created workload identity provider.
            """

            if 'attributesMap' in kwargs.keys():
                kwargs['attributesMap'] = self._build_attributes_map_list(attributes_map=kwargs['attributesMap'])

            return self.britive.post(self.base_url, json=kwargs)

        def create_aws(self, name: str, attributes_map: dict, description: str = None,
                       validation_window: int = 30, max_duration: int = 5) -> dict:
            """
            Create an AWS workload identity provider.

            :param name: Name of the AWS workload identity provider.
            :param attributes_map: The mapping of the available AWS workload token fields to custom identity attributes
                which are associated with service identities. Defaults to None.Can provide either the name of the
                custom identity attribute or the ID.
            :param description: Optional description of the AWS workload identity provider.
            :param validation_window: The number of seconds allowed to validate the AWS workload token from its
                "issued time". Defaults to 30 seconds.
            :param max_duration: The max number of hours (whole numbers only) for which the AWS workload token is valid.
            :returns: Details of the newly created AWS workload identity provider.
            """

            params = {
                'name': name,
                'idpType': 'AWS',
                'validationWindow': validation_window,
                'maxDuration': max_duration
            }
            if description:
                params['description'] = description
            if attributes_map:
                params['attributesMap'] = attributes_map

            return self.create(**params)

        def create_oidc(self, name: str, issuer_url: str, attributes_map: dict = None, description: str = None,
                        validation_window: int = 30, allowed_audiences: list = None) -> dict:
            """
            Create an OIDC workload identity provider.

            :param name: Name of the OIDC workload identity provider.
            :param issuer_url: The issuer url for the OIDC provider.
            :param attributes_map: The mapping of the available OIDC workload token fields to custom identity attributes
                which are associated with service identities. Defaults to None. Can provide either the name of the
                custom identity attribute or the ID.
            :param description: Optional description of the AWS workload identity provider.
            :param validation_window: The number of seconds allowed to validate the AWS workload token from its
                "issued time". Defaults to 30 seconds.
            :param allowed_audiences: The list of allowed audience values as strings. Defaults to None.
            :returns: Details of the newly created OIDC workload identity provider.
            """

            params = {
                'name': name,
                'idpType': 'OIDC',
                'validationWindow': validation_window,
                'issuerUrl': issuer_url
            }
            if description:
                params['description'] = description
            if allowed_audiences:
                params['allowedAudiences'] = allowed_audiences
            if attributes_map:
                params['attributesMap'] = attributes_map

            return self.create(**params)

        def update(self, workload_identity_provider_id: int, **kwargs) -> dict:
            """
            Updates a workload identity provider.

            This method accepts a list of kwargs which match the required and optional fields for creating a
            workload identity provider. It mostly exists to support the update of new workload identity provider
            types before this SDK can be updated to support the new type natively.

            Generally, the caller should opt to use the `update_aws` or `update_oidc` methods instead of calling
            this method directly.

            The field `attributesMap` is in format `{'idp_attr_value': 'custom_identity_attribute_name_or_id', ...}`.

            :param workload_identity_provider_id: The ID of the workload identity provider to update.
            :param kwargs: A list of keyword arguments which will be provided directly to the API backend without
                any further inspection. Valid fields are `name`, `description`, `idpType`, `attributesMap`, and
                `validationWindow`. For the AWS provider an additional field `maxDuration` is valid. For the OIDC
                provider additional fields `issuerUrl` and `allowedAudiences` are valid.
            :returns: Details of the updated workload identity provider.
            """

            kwargs['id'] = workload_identity_provider_id

            if 'attributesMap' in kwargs.keys():
                kwargs['attributesMap'] = self._build_attributes_map_list(attributes_map=kwargs['attributesMap'])

            # since this is a PUT call and not a PATCH call we need to get the existing idp configuration
            # and merge in the things that have changed

            existing = self.get(workload_identity_provider_id=workload_identity_provider_id)
            return self.britive.put(self.base_url, json={**existing, **kwargs})

        def update_aws(self, workload_identity_provider_id: int, name: str = None, attributes_map: dict = None,
                       description: str = None, validation_window: int = None, max_duration: int = None) -> dict:
            """
            Update an AWS workload identity provider.

            All fields except `workload_identity_provider_id` are optional.

            :param workload_identity_provider_id: The ID of the workload identity provider to update.
            :param name: Name of the AWS workload identity provider.
            :param attributes_map: The mapping of the available AWS workload token fields to custom identity attributes
                which are associated with service identities.Can provide either the name of the custom identity
                attribute or the ID.
            :param description: Description of the AWS workload identity provider.
            :param validation_window: The number of seconds allowed to validate the AWS workload token from its
                "issued time".
            :param max_duration: The max number of hours (whole numbers only) for which the AWS workload token is valid.
            :returns: Details of the updated AWS workload identity provider.
            """

            params = {
                'idpType': 'AWS'
            }

            if name:
                params['name'] = name
            if description:
                params['description'] = description
            if validation_window:
                params['validationWindow'] = validation_window
            if max_duration:
                params['maxDuration'] = max_duration
            if attributes_map:
                params['attributesMap'] = attributes_map

            return self.update(workload_identity_provider_id=workload_identity_provider_id, **params)

        def update_oidc(self, workload_identity_provider_id: int, name: str = None, issuer_url: str = None,
                        attributes_map: dict = None, description: str = None, validation_window: int = None,
                        allowed_audiences: list = None) -> dict:
            """
            Update an OIDC workload identity provider.

            All fields except `workload_identity_provider_id` are optional.

            :param workload_identity_provider_id: The ID of the workload identity provider to update.
            :param name: Name of the OIDC workload identity provider.
            :param issuer_url: The issuer url for the OIDC provider.
            :param attributes_map: The mapping of the available OIDC workload token fields to custom identity attributes
                which are associated with service identities. Can provide either the name of the custom identity
                attribute or the ID.
            :param description: Description of the AWS workload identity provider.
            :param validation_window: The number of seconds allowed to validate the AWS workload token from its
                "issued time".
            :param allowed_audiences: The list of allowed audience values as strings.
            :returns: Details of the update OIDC workload identity provider.
            """

            params = {
                'idpType': 'OIDC'
            }

            if name:
                params['name'] = name
            if description:
                params['description'] = description
            if validation_window:
                params['validationWindow'] = validation_window
            if issuer_url:
                params['issuerUrl'] = issuer_url
            if attributes_map:
                params['attributesMap'] = attributes_map
            if allowed_audiences:
                params['allowedAudiences'] = allowed_audiences

            return self.update(workload_identity_provider_id=workload_identity_provider_id, **params)

        def delete(self, workload_identity_provider_id) -> None:
            """
            Deletes a workload identity provider.

            :param workload_identity_provider_id: The ID of the workload identity provider.
            :returns: None.
            """
            return self.britive.delete(f'{self.base_url}/{workload_identity_provider_id}')

        def generate_attribute_map(self, idp_attribute_name: str, custom_identity_attribute_name: str = None,
                                   custom_identity_attribute_id: str = None) -> dict:
            """
            Generates a dictionary that can be appended to a list used for the `attributesMap`.

            This method would mostly be used when invoking the `create` or `update` methods directly instead of
            using the type specific (`create_aws`, `create_oidc`, `update_aws`, `update_oidc`) methods which provide
            a more pythonic way to capture the attribute mappings.

            :param idp_attribute_name: The name of the workload identity provider attribute to map. This will always
                be a string as it is controlled by the identity provider.
            :param custom_identity_attribute_name: The name of the Britive custom identity attribute. One of
                `custom_identity_attribute_name` or `custom_identity_attribute_id` must be provided. The name will be
                translated to the ID of the custom identity attribute on behalf of the caller.
            :param custom_identity_attribute_id: The id of the Britive custom identity attribute. One of
                `custom_identity_attribute_name` or `custom_identity_attribute_id` must be provided.
            :returns: Dictionary representing the attribute map.
            """

            if custom_identity_attribute_id and custom_identity_attribute_name:
                raise ValueError('only one of custom_identity_attribute_id and custom_identity_attribute_name '
                                 'should be provided')

            if not custom_identity_attribute_id and not custom_identity_attribute_name:
                raise ValueError('one of custom_identity_attribute_id or custom_identity_attribute_name '
                                 'should be provided')

            if custom_identity_attribute_name:
                found = False
                for attr in self.britive.identity_attributes.list():
                    if attr['name'] == custom_identity_attribute_name:
                        custom_identity_attribute_id = attr['id']
                        found = True
                        break
                if not found:
                    raise ValueError(f'custom_identity_attribute_name value of {custom_identity_attribute_name} '
                                     f'not found.')

            return {
                'idpAttr': idp_attribute_name,
                'userAttr': custom_identity_attribute_id
            }

    class ServiceIdentities:
        def __init__(self, workload):
            self.britive = workload.britive
            self.base_url: str = workload.base_url + '/users/{id}/identity-provider'  # will .format(id=...) later

        def get(self, service_identity_id: str) -> Union[dict, None]:
            """
            Returns details about the workload identity provider associated with the specified service identity.

            :param service_identity_id: The ID of the service identity.
            :returns: Details about the workload identity provider associated with the specified service identity.
                Returns None if there is no workload identity provider associated with the specified service identity.
            """

            return self.britive.get(self.base_url.format(id=service_identity_id))

        def assign(self, service_identity_id: str, idp_id: str, federated_attributes: dict,
                   token_duration: int = 300) -> dict:
            """
            Associates an OIDC provider with the specified Service Identity.

            :param service_identity_id: The ID of the service identity.
            :param idp_id: The ID of the OIDC Identity Provider.
            :param token_duration: Duration in seconds (from now) before a token provided by the client expires.
                This will be evaluated alongside the OIDC JWT expiration field and the earlier of the two values
                will govern when the token expires. This field is optional and defaults to 300 seconds.
            :param federated_attributes: An attribute map where keys are the custom attribute ids or names
                and values are strings or list of strings (for multivalued attributes) which map back to the mapped
                token claims.
            :returns: Details of the newly assigned workload identity provider.
            """

            mapping_attributes = []
            converted_federated_attributes = {}
            converted_attributes = self.britive.service_identities.custom_attributes._build_list(
                operation='add',
                custom_attributes=federated_attributes
            )

            for attr in converted_attributes:
                custom_attr = attr['customUserAttribute']
                attr_id = custom_attr['attributeId']
                if attr_id not in converted_federated_attributes.keys():
                    converted_federated_attributes[attr_id] = []
                converted_federated_attributes[attr_id].append(custom_attr['attributeValue'])

            for custom_attribute_id, values in converted_federated_attributes.items():
                mapping_attributes.append(
                    {
                        'attrId': custom_attribute_id,
                        'values': values
                    }
                )

            params = {
                'idpId': idp_id,
                'tokenDuration': token_duration,
                'mappingAttributes': mapping_attributes
            }

            return self.britive.post(self.base_url.format(id=service_identity_id), json=params)

        def unassign(self, service_identity_id: str) -> None:
            """
            Removes/deletes the service identity's assigned identity provider, along with any custom attribute mappings.

            This will revert the service identity back to use static API tokens for authentication.

            :param service_identity_id: The ID of the Service Identity.
            :returns: None.
            """
            return self.britive.delete(self.base_url.format(id=service_identity_id))

    class ScimUser:
        def __init__(self, workload):
            self.britive = workload.britive
            self.base_url: str = workload.base_url + '/scim-user/identity-provider'

        def get(self, idp_name: str) -> dict:
            """
            Gets details of the workload federation enabled service identity associated with the identity provider.

            The identity provider name provided must be a SAML identity provider.

            :param idp_name: The name of the SAML Identity Provider.
            :returns: Details of the newly assigned service identity to the identity provider.
            """

            return self.britive.get(f'{self.base_url}/{idp_name}')

        def assign(self, service_identity_id: str, idp_name: str) -> dict:
            """
            Associates a workload federation enabled service identity with an identity provider SCIM service.

            The service identity must already be configured for workload federation.
            The identity provider name provided must be a SAML identity provider.

            :param service_identity_id: The ID of the service identity.
            :param idp_name: The name of the SAML Identity Provider.
            :returns: Details of the newly assigned service identity to the identity provider.
            """

            params = {
                'idpName': idp_name,
                'userId': service_identity_id
            }

            return self.britive.post(self.base_url, json=params)

        def unassign(self, idp_name: str) -> dict:
            """
            Removes a workload federation enabled service identity from an identity provider SCIM service.

            The identity provider name provided must be a SAML identity provider.

            :param idp_name: The name of the SAML Identity Provider.
            :returns: Details of the newly assigned service identity to the identity provider.
            """

            return self.britive.delete(f'{self.base_url}/{idp_name}')




class AccessDenied(Exception):
    pass


class ApiTokenNotFound(Exception):
    pass


class ApprovalRequiredButNoJustificationProvided(Exception):
    pass


class ApprovalWorkflowRejected(Exception):
    pass


class ApprovalWorkflowTimedOut(Exception):
    pass


class ForbiddenRequest(Exception):
    pass


class InternalServerError(Exception):
    pass


class InvalidFederationProvider(Exception):
    pass


class InvalidRequest(Exception):
    pass


class MethodNotAllowed(Exception):
    pass


class NoSecretsVaultFound(Exception):
    pass


class NotExecutingInAzureEnvironment(Exception):
    pass


class NotExecutingInBitbucketEnvironment(Exception):
    pass


class NotExecutingInGithubEnvironment(Exception):
    pass


class NotExecutingInSpaceliftEnvironment(Exception):
    pass


class NotFound(Exception):
    pass


class ProfileApprovalMaxBlockTimeExceeded(Exception):
    pass


class ProfileApprovalRejected(Exception):
    pass


class ProfileApprovalTimedOut(Exception):
    pass


class ProfileApprovalWithdrawn(Exception):
    pass


class ProfileCheckoutAlreadyApproved(Exception):
    pass


class ProfileNotFound(Exception):
    pass


class RootEnvironmentGroupNotFound(Exception):
    pass


class ServiceUnavailable(Exception):
    pass


class TenantMissingError(Exception):
    pass


class TokenMissingError(Exception):
    pass


class TooManyServiceIdentitiesFound(Exception):
    pass


class TooManyUsersFound(Exception):
    pass


class TransactionNotFound(Exception):
    pass


class UnauthorizedRequest(Exception):
    pass


class UserDoesNotHaveMFAEnabled(Exception):
    pass


class UserNotAllowedToChangePassword(Exception):
    pass


class UserNotAssociatedWithDefaultIdentityProvider(Exception):
    pass


class TenantUnderMaintenance(Exception):
    pass


# from https://docs.britive.com/docs/restapi-status-codes
allowed_exceptions = {
    400: InvalidRequest,
    401: UnauthorizedRequest,
    403: ForbiddenRequest,
    404: NotFound,
    405: MethodNotAllowed,
    500: InternalServerError,
    503: ServiceUnavailable
}



from .cache import *  # will also import some globals like `britive`


def test_create(cached_notification):
    assert isinstance(cached_notification, dict)
    assert cached_notification['description'] == 'test'


def test_list():
    notifications = britive.notifications.list()
    assert isinstance(notifications, list)
    assert len(notifications) > 0


def test_get(cached_notification):
    notification = britive.notifications.get(notification_id=cached_notification['notificationId'])
    assert isinstance(notification, dict)
    assert notification['notificationId'] == cached_notification['notificationId']


def test_update(cached_notification):
    notification = britive.notifications.update(
        notification_id=cached_notification['notificationId'],
        description='test2'
    )
    assert isinstance(notification, dict)
    assert notification['description'] == 'test2'


def test_disable(cached_notification):
    notification = britive.notifications.disable(notification_id=cached_notification['notificationId'])
    assert isinstance(notification, dict)
    assert notification['status'] == 'Inactive'


def test_enable(cached_notification):
    notification = britive.notifications.enable(notification_id=cached_notification['notificationId'])
    assert isinstance(notification, dict)
    assert notification['status'] == 'Active'


def test_available_rules(cached_notification_rules):
    assert isinstance(cached_notification_rules, list)


def test_available_users(cached_notification_users):
    assert isinstance(cached_notification_users, list)


def test_available_user_tags(cached_notification_user_tags):
    assert isinstance(cached_notification_user_tags, list)


def test_available_applications(cached_notification_applications):
    assert isinstance(cached_notification_applications, list)


def test_configure(cached_notification, cached_notification_rules, cached_notification_applications, cached_user):
    rules = []
    for rule in cached_notification_rules:
        if rule['predicate'] in ['AccountsCreated', 'AccountsDeleted']:
            rules.append(rule)

    response = britive.notifications.configure(
        notification_id=cached_notification['notificationId'],
        users=[cached_user['userId']],
        rules=rules,
        send_no_changes=True
    )

    assert isinstance(response, dict)
    assert response['sendNoChanges']
    assert len(response['memberRules']) == 1
    assert len(response['rules']) == 2


def test_delete(cached_notification):
    response = britive.notifications.delete(notification_id=cached_notification['notificationId'])
    assert response is None
    cleanup('notification')
    cleanup('notification-available-rules')
    cleanup('notification-available-users')
    cleanup('notification-available-user-tags')
    cleanup('notification-available-applications')




from .cache import *  # will also import some globals like `britive`
import json


def test_whoami():
    me = britive.my_access.whoami()
    assert isinstance(me, dict)
    print(json.dumps(me, indent=2, default=str))


def test_list_profiles():
    profiles = britive.my_access.list_profiles()
    assert isinstance(profiles, list)
    assert len(profiles) > 0
    assert isinstance(profiles[0], dict)


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_checkout(cached_checked_out_profile):
    assert isinstance(cached_checked_out_profile, dict)
    assert cached_checked_out_profile['accessType'] == 'PROGRAMMATIC'
    assert 'credentials' in cached_checked_out_profile.keys()


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_checkout_again(cached_profile, cached_environment):
    response = britive.my_access.checkout(
            profile_id=cached_profile['papId'],
            environment_id=cached_environment['id'],
            include_credentials=True
        )

    assert isinstance(response, dict)
    assert response['accessType'] == 'PROGRAMMATIC'
    assert 'credentials' in response.keys()


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_list_checked_out_profiles():
    profiles = britive.my_access.list_checked_out_profiles()
    assert isinstance(profiles, list)
    assert len(profiles) >= 1  # since we just checked one out!


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_get_checked_out_profile(cached_checked_out_profile):
    profile = britive.my_access.get_checked_out_profile(transaction_id=cached_checked_out_profile['transactionId'])
    assert isinstance(profile, dict)


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_checkin(cached_checked_out_profile):
    response = britive.my_access.checkin(transaction_id=cached_checked_out_profile['transactionId'])
    assert isinstance(response, dict)
    assert response['status'] in ('checkedIn', 'checkInSubmitted')  # v1, v2
    cleanup('checked-out-profile')


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_checkout_by_name(cached_checked_out_profile_by_name):
    assert isinstance(cached_checked_out_profile_by_name, dict)
    assert cached_checked_out_profile_by_name['accessType'] == 'PROGRAMMATIC'
    assert 'credentials' in cached_checked_out_profile_by_name.keys()


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_checkin_by_name(cached_profile, cached_environment, cached_application):
    account_id = os.environ['BRITIVE_TEST_ENV_ACCOUNT_ID']
    response = britive.my_access.checkin_by_name(
        profile_name=cached_profile['name'],
        environment_name=f'{account_id} ({cached_environment["name"]})',
        application_name=cached_application['catalogAppDisplayName']
    )
    assert isinstance(response, dict)
    assert response['status'] in ('checkedIn', 'checkInSubmitted')  # v1, v2
    cleanup('checked-out-profile-by-name')


def test_frequents():
    profiles = britive.my_access.frequents()
    assert isinstance(profiles, list)


def test_favorites():
    profiles = britive.my_access.favorites()
    assert isinstance(profiles, list)


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_request_and_approve(cached_profile, cached_service_identity_token, cached_environment,
                             cached_service_identity):

    token = britive.service_identity_tokens.create(cached_service_identity['userId'], 90)['token']
    other_britive = Britive(token=token, query_features=False)

    request = other_britive.my_access.request_approval(
        profile_id=cached_profile['papId'],
        environment_id=cached_environment['id'],
        justification='let me in'
    )

    assert 'requestId' in request.keys()

    request_id = request['requestId']

    approvals = britive.my_access.list_approvals()
    for approval in approvals:
        if approval['requestId'] == request_id:
            assert approval['status'] == 'PENDING'
            break

    response = britive.my_access.reject_request(
        request_id=request_id
    )

    assert response is None

    approvals = britive.my_access.list_approvals()
    for approval in approvals:
        if approval['requestId'] == request_id:
            assert approval['status'] == 'REJECTED'
            break

    request = other_britive.my_access.request_approval(
        profile_id=cached_profile['papId'],
        environment_id=cached_environment['id'],
        justification='let me in'
    )

    assert 'requestId' in request.keys()

    request_id = request['requestId']

    approvals = britive.my_access.list_approvals()
    for approval in approvals:
        if approval['requestId'] == request_id:
            assert approval['status'] == 'PENDING'
            break

    response = britive.my_access.approve_request(
        request_id=request_id
    )

    assert response is None

    approvals = britive.my_access.list_approvals()
    for approval in approvals:
        if approval['requestId'] == request_id:
            assert approval['status'] == 'APPROVED'
            break




from .cache import *  # will also import some globals like `britive`


def test_catalog(cached_catalog):
    assert isinstance(cached_catalog, dict)
    assert 'AWS Standalone-1.0' in cached_catalog.keys()
    assert isinstance(cached_catalog['AWS Standalone-1.0'], dict)


def test_create(cached_application):
    show = None
    for prop in cached_application['catalogApplication']['propertyTypes']:
        if prop['name'] == 'showAwsAccountNumber':
            show = prop['value']
            break
    assert isinstance(cached_application, dict)
    assert show is False
    assert isinstance(cached_application['userAccountMappings'], list)
    assert len(cached_application['userAccountMappings']) == 0


def test_disable(cached_application):
    response = britive.applications.disable(application_id=cached_application['appContainerId'])
    assert isinstance(response, dict)
    assert response['status'] == 'inactive'


def test_enable(cached_application):
    response = britive.applications.enable(application_id=cached_application['appContainerId'])
    assert isinstance(response, dict)
    assert response['status'] == 'active'


def test_list(cached_application):
    apps = britive.applications.list()
    assert isinstance(apps, list)
    assert len(apps) > 0
    assert isinstance(apps[0], dict)
    assert cached_application['appContainerId'] in [a['appContainerId'] for a in apps]


def test_get(cached_application):
    app = britive.applications.get(application_id=cached_application['appContainerId'])
    assert app['appContainerId'] == cached_application['appContainerId']
    assert app['catalogAppDisplayName'] == cached_application['catalogAppDisplayName']


def test_test_failure(cached_application):
    response = britive.applications.test(application_id=cached_application['appContainerId'])
    assert isinstance(response, dict)
    assert 'success' in response.keys()
    assert 'message' in response.keys()
    assert not response['success']


def test_update(cached_application):
    tenant = britive.tenant.replace('.britive-app.com', '')
    idp = os.environ.get('BRITIVE_IDP_NAME_OVERRIDE') or f'BritivePythonApiWrapperTesting-{tenant}'
    role = os.environ.get('BRITIVE_INTEGRATION_ROLE_NAME_OVERRIDE') or f'britive-integration-role-{tenant}'

    app = britive.applications.update(
        application_id=cached_application['appContainerId'],
        showAwsAccountNumber=True,
        identityProvider=idp,
        roleName=role
    )
    show = None
    for prop in app['catalogApplication']['propertyTypes']:
        if prop['name'] == 'showAwsAccountNumber':
            show = prop['value']
            break
    assert isinstance(app, dict)
    assert show is True


def test_set_user_account_mapping(cached_application):
    app = britive.applications.set_user_account_mapping(cached_application['appContainerId'], 'email')
    assert isinstance(app['userAccountMappings'], list)
    assert len(app['userAccountMappings']) == 1
    assert app['userAccountMappings'][0]['name'] == 'email'




from .cache import *  # will also import some globals like `britive`


def test_fields():
    fields = britive.audit_logs.fields()
    assert isinstance(fields, dict)
    assert len(fields.keys()) == 18


def test_operators():
    operators = britive.audit_logs.operators()
    assert isinstance(operators, dict)
    assert len(operators.keys()) == 3


def test_query_json():
    events = britive.audit_logs.query()
    assert isinstance(events, list)
    assert isinstance(events[0], dict)
    assert len(events) % 100 != 0  # v2.8.1 - adding check due to pagination bug not including the last page


def test_query_csv():
    csv = britive.audit_logs.query(csv=True)
    assert '"timestamp","actor.display_name"' in csv






class EnvironmentGroups:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/apps'

    def create(self, application_id: str, name: str, description: str = None, parent_id: str = None) -> dict:
        """
        Create a new environment group inside the specified application.

        Only applicable to applications which have `catalogApplication.supportsEnvironmentScanning` set to `False`.

        :param application_id: The ID of the application inside of which the environment group will be created.
        :param name: The name of the environment group.
        :param description: An optional description of the environment group.
        :param parent_id: An optional parent group id - if not provided then the root group is assumed and will
            be created as needed.
        :return: Details of the newly created environment group.
        """

        data = {
            'name': name,
            'type': 'group',
            'description': description or '',
            'parentId': parent_id or self.get_or_create_root(application_id=application_id)
        }

        return self.britive.post(f'{self.base_url}/{application_id}/root-environment-group/groups', json=data)

    def get_or_create_root(self, application_id: str) -> str:
        """
        Returns the ID of the root environment group. Will only create one if one does not already exist.

        A root environment group is NOT automatically created at application creation. For applications which
        support environments and environment groups (`catalogApplication.supportsEnvironmentScanning` set to `False`)
        the root group has to be created before the first environment/environment group can be created. That will be
        handled by this method and is internal to this SDK - the end user would generally not need to  interact with
        this method. Instead the logic inside of EnvironmentGroups.create() and Environments.create() handles the
        creation of the root group as required.

        :param application_id: The ID of the application under which to create the root environment group.
        :return: ID of the root environment group for the given application.
        """

        groups = self.list(application_id=application_id)
        root_id = None
        for group in groups:
            if group['name'].lower() == 'root' and group['parentId'] == '' and group['type'] == 'group':
                root_id = group['id']
                break

        if not root_id:
            data = {
                'name': 'root',
                'type': 'group',
                'description': '',
                'parentId': ''
            }

            root_id = self.britive.post(
                f'{self.base_url}/{application_id}/root-environment-group/groups',
                json=data
            )['id']

        return root_id

    def get(self, application_id: str, environment_group_id: str) -> dict:
        """
        Return details about the specified environment group.

        :param application_id: The ID of the application where the environment group resides.
        :param environment_group_id: The ID of the environment group.
        :return: Details about the specified environment group.
        """

        groups = self.list(application_id=application_id)
        for group in groups:
            if group['id'] == environment_group_id:
                return group
        return {}

    def list(self, application_id: str) -> list:
        """
        Return a list of environment groups for the specified application.

        :param application_id: The ID of the application.
        :return: List of environment groups.
        """

        app = self.britive.get(f'{self.base_url}/{application_id}')
        root_env_group = app.get('rootEnvironmentGroup', {}) or {}
        groups = root_env_group.get('environmentGroups')
        return groups or []

    def delete(self, application_id: str, environment_group_id: str) -> None:
        """
        Delete the specified environment group.

        :param application_id: The ID of the application containing the environment group to delete.
        :param environment_group_id: The ID of the environment group to delete.
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/{application_id}/environment-groups/{environment_group_id}')






class Notifications:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/notifications'

    def list(self) -> list:
        """
        List notifications.

        :return: List of notifications.
        """

        return self.britive.get(self.base_url)

    def get(self, notification_id: str) -> dict:
        """
        Retrieve details about a notification.

        :param notification_id: The ID of the notification.
        :return: Details of the notification
        """

        return self.britive.get(f'{self.base_url}/{notification_id}')

    def create(self, name: str, description: str = None) -> dict:
        """
        Create a new notification.

        :param name: The name of the notification.
        :param description: An optional description of the notification.
        :return: Details of the newly created notification.
        """

        data = {
            'name': name,
            'description': description or ''
        }
        return self.britive.post(self.base_url, json=data)

    def update(self, notification_id: str, name: str = None, description: str = None) -> dict:
        """
        Update specific attributes of a notification.

        :param notification_id: The ID of the notification.
        :param name: The optional new name of the notification. Omitting will leave the name unchanged.
        :param description: The optional new description of the notification. Omitting will leave the description
            unchanged.
        :return: Details of the updated notification.
        """

        base = self.get(notification_id=notification_id)
        if name:
            base['name'] = name
        if description:
            base['description'] = description
        return self.britive.put(self.base_url, json=base)

    def available_rules(self) -> list:
        """
        List all supported notification rules.

        :return: List of notification rules which can be used to craft the events on which to notify.
        """

        return self.britive.get(f'{self.base_url}/supported-rules')

    def available_users(self, notification_id: str) -> list:
        """
        Provide list of available users which can be added to the notification.

        :param notification_id: The ID of the notification.
        :return: List of available users for the notification.
        """

        return self.britive.get(f'{self.base_url}/{notification_id}/available-users')

    def available_user_tags(self, notification_id: str) -> list:
        """
        Provide list of available user tags which can be added to the notification.

        :param notification_id: The ID of the notification.
        :return: List of available user tags for the notification.
        """

        return self.britive.get(f'{self.base_url}/{notification_id}/available-user-tags')

    def available_applications(self, notification_id: str) -> list:
        """
        Provide list of available applications which can be added to the notification.

        :param notification_id: The ID of the notification.
        :return: List of available applications for the notification.
        """

        return self.britive.get(f'{self.base_url}/{notification_id}/available-apps')

    def configure(self, notification_id: str, rules: list = None, users: list = None, user_tags: list = None,
                  applications: list = None, send_no_changes: bool = None, notification_medium_id: str = None) -> dict:
        """
        Configure the details of a notification.

        For all optional parameters omitting the parameter will leave the value unchanged.

        :param notification_id: The ID of the notification.
        :param rules: List of rules to apply. Obtain rule options from `britive.notifications.available_rules()` and
            use results from that API call to populate this list. Maximum of 3 rules are allowed.
        :param users: List of user ids to apply. This is the list of users who will be notified if any of the rules are
            triggered. An empty list means that no users will be notified.
        :param user_tags: List of user tag ids to apply. This is the list of user tags who will be notified if any of
            the rules are triggered. An empty list means that no user tags will be notified.
        :param applications: List of applications to which this notification applies. Obtain applications options from
            `britive.notifications.available_applications()` and use results from that API call to populate this list.
            An empty list indicates the event applies to all applications.
        :param send_no_changes: Boolean indicating whether to send notification regardless of whether any changes have
            occurred or not.
        :param notification_medium_id: The ID of the notification medium to use for this notification.
        :return: Details of the newly updated notification.
        """

        # some basic validation
        if rules and len(rules) > 3:
            raise ValueError('The maximum number of rules for a notification is 3.')

        # start with the existing notification details
        data = self.get(notification_id=notification_id)

        members = []
        for user in self.britive.users.minimized_user_details(user_ids=users):
            members.append(
                {
                    'id': user['id'],
                    'memberType': 'User',
                    'name': user['username'],
                    'condition': None
                }
            )

        for tag in self.britive.tags.minimized_tag_details(tag_ids=user_tags):
            members.append(
                {
                    'id': tag['userTagId'],
                    'memberType': 'Tag',
                    'name': tag['name'],
                    'condition': None
                }
            )

        # set the possible parameters
        params = {
            'rules': rules,
            'memberRules': members,
            'applications': applications,
            'sendNoChanges': send_no_changes,
            'notificationMedium': notification_medium_id
        }

        # fo each parameter update the existing notification data if the parameter was provided
        for key, value in params.items():
            if value:
                data[key] = value

        return self.britive.put(self.base_url, json=data)

    def disable(self, notification_id: str) -> dict:
        """
        Disable a notification.

        :param notification_id: The ID of the notification.
        :return: Details of the newly disabled notification.
        """

        return self.britive.post(f'{self.base_url}/{notification_id}/disabled-statuses')

    def enable(self, notification_id: str) -> dict:
        """
        Enable a notification.

        :param notification_id: The ID of the notification.
        :return: Details of the newly enabled notification.
        """

        return self.britive.post(f'{self.base_url}/{notification_id}/enabled-statuses')

    def delete(self, notification_id: str) -> None:
        """
        Delete a notification.

        :param notification_id: The ID of the notification.
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/{notification_id}')



from .roles import SystemRoles
from .policies import SystemPolicies
from .permissions import SystemPermissions
from .consumers import SystemConsumers
from .actions import SystemActions


# this class is just a logical grouping construct


class System:
    def __init__(self, britive):
        self.britive = britive
        self.roles = SystemRoles(britive)
        self.policies = SystemPolicies(britive)
        self.permissions = SystemPermissions(britive)
        self.consumers = SystemConsumers(britive)
        self.actions = SystemActions(britive)




from .cache import *  # will also import some globals like `britive`
import os


def test_list():
    attributes = britive.identity_attributes.list()
    assert isinstance(attributes, list)
    assert len(attributes) >= 7  # there are 7 default attributes that are managed by the system


def test_create(cached_identity_attribute):
    assert isinstance(cached_identity_attribute, dict)





class SystemPermissions:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/v1/policy-admin/permissions'

    @staticmethod
    def _validate_identifier_type(identifier_type):
        if identifier_type not in ['id', 'name']:
            raise ValueError(f'identifier_type of {identifier_type} is invalid. Only `name` and `id` are allowed.')

    def list(self, filter_expression: str = '') -> list:
        """
        List system level permissions.

        :param filter_expression: Filter based on `name`. Valid operators are `eq`, `sw`, and
            `co`. Example: name co view
        :returns: List of permissions.
        """

        params = {}
        if filter_expression:
            params['filter'] = filter_expression
        return self.britive.get(self.base_url, params=params)

    def get(self, permission_identifier: str, identifier_type: str = 'name') -> dict:
        """
        Get details of the specified permission.

        :param permission_identifier: The ID or name of the permission.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers will be returned. Either all identifiers must be names or all identifiers must be IDs.
        :returns: Details of the specified permission.
        """

        self._validate_identifier_type(identifier_type)

        return self.britive.get(f'{self.base_url}/{permission_identifier}')

    def create(self, permission: dict) -> dict:
        """
        Create a system level permission.

        :param permission: The permission to create. Use `permissions.build` to assist in constructing a proper
            permission document.
        :returns: Details of the newly created permission.
        """

        return self.britive.post(self.base_url, json=permission)
    
    def update(self, permission_identifier: str, permission: dict, identifier_type: str = 'name') -> None:
        """
        Update a system level permission.

        :param permission_identifier: The ID or name of the permission to update.
        :param permission: The permission to update. Use `permissions.build` to assist in constructing a proper
            permission document.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers will be returned. Either all identifiers must be names or all identifiers must be IDs.
        :returns: None.
        """

        self._validate_identifier_type(identifier_type)

        if permission.get('isInline'):
            raise ValueError('attribute isInline is set to True - cannot update an inline permission.')

        permission.pop('isInline', None)  # InvalidRequest: 400 - PA-0059 - isInline is not allowed to update
        permission.pop('isReadOnly', None)
        return self.britive.patch(f'{self.base_url}/{permission_identifier}', json=permission)

    def delete(self, permission_identifier: str, identifier_type: str = 'name') -> None:
        """
        Delete a system level permission.

        :param permission_identifier: The ID or name of the permission to delete.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers will be returned. Either all identifiers must be names or all identifiers must be IDs.
        :returns: None.
        """

        self._validate_identifier_type(identifier_type)
        return self.britive.delete(f'{self.base_url}/{permission_identifier}')

    @staticmethod
    def build(name: str, consumer: str, actions: list, resources: list = [], description: str = '',
              read_only: bool = False, is_inline: bool = False) -> dict:
        """
        Build a permission document given the provided inputs.

        :param name: The name of the permission.
        :param consumer: List of permission names or ids this permission grants.
        :param actions: List of action names for the permission.
        :param resources: List of resources IDs for which to scope this permission. If omitted then `*` will be applied.
        :param description: An optional description of the permission.
        :param read_only: Indicates if the permission is a read only. Defaults to `False`.
        :param is_inline: Indicates if the permission is inline to a role/policy. Defaults to `False`. `False`
            indicates that the permission is its own entity which can be assigned to roles/polices. `True` indicates
            the permissions is unique/inline to an role/policy.
        :return: A dict which can be provided as a permission document to `create` and `update`.
        """

        # put it all together
        permission = {
            'name': name,
            'description': description,
            'isReadOnly': read_only,
            'isInline': is_inline,
            'consumer': consumer,
            'actions': actions,
            'resources': resources if len(resources) > 0 else ['*']
        }

        return permission



class IdentityProviders:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/identity-providers'
        self.scim_tokens = ScimTokens(britive)
        self.scim_attributes = ScimAttributes(britive)

    def list(self) -> list:
        """
        Return list of identity providers.

        :return: List of identity providers
        """

        return self.britive.get(self.base_url)

    def get(self, identity_provider_id: str) -> dict:
        """
        Return details of the specified identity provider.

        :param identity_provider_id: The ID of the identity provider.
        :return: Details of the identity provider.
        """

        return self.britive.get(f'{self.base_url}/{identity_provider_id}')

    def get_by_id(self, identity_provider_id: str) -> dict:
        """
        Return details of the given identity provider.

        Equivalent to `get()`. Exists for parity with `get_by_name` since identity providers can be found by
        `name` in addition to `id`.
        :param identity_provider_id:
        :return: Details of the identity provider.
        """

        return self.get(identity_provider_id=identity_provider_id)

    def get_by_name(self, identity_provider_name: str) -> dict:
        """
        Return details of the given identity provider.

        :param identity_provider_name: Name of the identity provider.
        :return: Details of the identity provider.
        """

        params = {
            'name': identity_provider_name
        }
        return self.britive.get(self.base_url, params=params)

    def signing_certificate(self) -> str:
        """
        Return the signing certificate for the Britive tenant.

        This certificate would be used to configure a new SSO identity provider. The certificate is the same
        across all identity providers in the Britive tenant.

        It is left to the caller to persist the certificate to disk, if required. File extension is `.cer`.

        :return: The signing certificate.
        """

        return self.britive.get(f'{self.base_url}/signing-certificate')

    def create(self, name: str, description: str = None) -> dict:
        data = {
            'name': name,
            'description': description or ''
        }
        return self.britive.post(self.base_url, json=data)

    def update(self, identity_provider_id: str, name: str = None, description: str = None, sso_provider: str = None,
               scim_provider: str = None) -> None:
        """
        Updates an identity provider.

        If `Azure` is used for SSO and SCIM provisioning, both SSO and SCIM providers need to be set to `Azure`.

        :param identity_provider_id: The ID of the identity provider.
        :param name: The name of the identity provider. If omitted will default to existing name.
        :param description: The description of the identity provider. If omitted will default to existing description.
        :param sso_provider: The SSO provider. Valid values are...

            - Generic
            - Azure

            If omitted will default to existing SSO provider.
        :param scim_provider: The SCIM provider. Valid values are...

            - Generic
            - Azure

            If omitted will default to existing SCIM provider.
        :return: None
        """

        # collect what needs to be updated
        possible_updates = {
            'name': name,
            'description': description,
            'scimProvider': scim_provider,
            'ssoProvider': sso_provider
        }
        updates = {k: v for k, v in possible_updates.items() if v is not None}

        # collect what already exists
        idp = self.get(identity_provider_id=identity_provider_id)
        existing = {
            'name': idp['name'],
            'description': idp['description'],
            'scimProvider': idp['scimProvider'],
            'ssoProvider': idp['ssoProvider']
        }

        # merge in any updates
        data = {**existing, **updates}

        # and finally perform the update
        return self.britive.patch(f'{self.base_url}/{identity_provider_id}', json=data)

    def delete(self, identity_provider_id: str) -> None:
        """
        Delete an identity provider.

        Performing this action WILL result in any remaining users associated with the identity provider losing access
        to the tenant.

        :param identity_provider_id: The ID of the identity provider.
        :return: None
        """

        self.britive.delete(f'{self.base_url}/{identity_provider_id}')

    def configure_mfa(self, identity_provider_id: str, root_user: bool = None, non_root_user: bool = None) -> None:
        """
        Enable or disable both root user multi factor authentication (MFA) and/or non root user MFA.

        :param identity_provider_id: The ID of the identity provider.
        :param root_user: Whether to enable or disable root user MFA. If omitted will default to existing configuration.
        :param non_root_user: Whether to enable or disable non root user MFA. If omitted will default to existing
            configuration.
        :return: None
        """

        # collect what needs to be updated
        possible_updates = {
            'mfaEnabled': non_root_user,
            'mfaEnabledRootUser': root_user
        }
        updates = {k: v for k, v in possible_updates.items() if v is not None}

        # collect what already exists
        idp = self.get(identity_provider_id=identity_provider_id)
        existing = {
            'mfaEnabled': idp['mfaEnabled'],
            'mfaEnabledRootUser': idp['mfaRootUserEnabled']
        }

        # merge in any updates
        data = {**existing, **updates}

        return self.britive.patch(f'{self.base_url}/{identity_provider_id}/mfa', json=data)

    def set_metadata(self, identity_provider_id: str, metadata_xml: str) -> dict:
        """
        Set SAML metadata for the specified identity provider.

        This metadata is in XML format, provided a a string, and would be obtained from the identity provider.

        :param identity_provider_id: The ID of the identity provider.
        :param metadata_xml: An XML string (str) representing SAML metadata obtained from the identity provider.
        :return: Details of the SAML metadata that was uploaded.
        """
        return self.britive.patch_upload(
            url=f'{self.base_url}/{identity_provider_id}/saml-metadata',
            file_content_as_str=metadata_xml,
            filename='file',
            content_type='text/xml'
        )


class ScimTokens:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/identity-providers'

    def create(self, identity_provider_id: str, token_expiration_days: int = 90) -> dict:
        """
        Create a SCIM token associated with the specified identity provider.

        Once created, any old token associated with the identity provider will be removed. Any identity provider can
        have only one SCIM token at any given time.

        The token that is generated will only be returned in a response once. It is the responsibility of the caller
        to persist this token.

        :param identity_provider_id: The ID of the identity provider.
        :param token_expiration_days: The number of days in which token would expire since it was last used.
            Value values are between 1 and 90.
        :return: Details of the newly created SCIM token.
        """

        data = {
            'tokenExpirationDays': token_expiration_days
        }
        return self.britive.post(f'{self.base_url}/{identity_provider_id}/scim-token', json=data)

    def get(self, identity_provider_id: str) -> dict:
        """
        Return details of SCIM token associated with the specified identity provider.

        :param identity_provider_id: The ID of the identity provider.
        :return: Details of the SCIM token.
        """

        return self.britive.get(f'{self.base_url}/{identity_provider_id}/scim-token')

    def update(self, identity_provider_id: str, token_expiration_days: int) -> None:
        """
        Update the token expiration days for the specified identity provider.

        :param identity_provider_id: The ID of the identity provider.
        :param token_expiration_days: The number of days in which token would expire since it was last used.
            Value values are between 1 and 90.
        :return: None
        """

        data = {
            'tokenExpirationDays': token_expiration_days
        }
        return self.britive.patch(f'{self.base_url}/{identity_provider_id}/scim-token', json=data)


class ScimAttributes:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/identity-providers'

    def list(self) -> list:
        """
        Return a list of SCIM attributes used for provisioning from the target system.

        :return: List of SCIM attributes.
        """

        return self.britive.get(f'{self.base_url}/scim-attributes')

    def update_mapping(self, identity_provider_id: str, mappings: list) -> None:
        """
        Add/remove SCIM mappings for the specified identity provider.

        When new SCIM mappings are added, the existing SCIM mappings for the identity attribute are deleted, if they
        are already defined.
        :param identity_provider_id: The ID of the identity provider.
        :param mappings: A list of identity attributes to add/remove. Each element in the list is a dict of the
            following format.

                {
                    "scimAttributeName": "addresses[type eq \"home\"].country",
                    "builtIn": False|True,
                    "attributeId": "HSvfZ1b74HUd7o8ai5Wo",
                    "attributeName": "country",
                    "op": "add"|"remove"
                }

            Each key is explained below in further detail.

            - scimAttributeName: The name of the SCIM attribute returned by
                `britive.identity_providers.scim_attributes.list()`.
            - builtIn: True if the identity attribute is from the built-in list, False if user generated.
            - attributeId: The ID of the identity attribute to be mapped with the identity provider. This can be
                obtained by calling `britive.identity_attributes.list()`.
            - attributeName: The name of the identity attribute to be mapped with the identity provider. This can be
                obtained by calling `britive.identity_attributes.list()`.
            - op: The operation to perform. Valid values are `add` and `remove`.
        :return: None
        """

        return self.britive.patch(f'{self.base_url}/{identity_provider_id}/scim-attribute-mappings', json=mappings)



import json
from typing import Union


class SecretsManager:
    def __init__(self, britive):
        self.vaults = Vaults(britive)
        self.password_policies = PasswordPolicies(britive)
        self.secrets = Secrets(britive)
        self.policies = Policies(britive)
        self.static_secret_templates = StaticSecretTemplates(britive)
        self.resources = Resources(britive)
        self.folders = Folders(britive)


class Vaults:
    def __init__(self, britive) -> None:
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/v1/secretmanager/vault'

    def list(self) -> list:
        """
        Provide a list of all vaults.

        :return: List of all vaults
        """

        params = {'getmetadata': 'true'}
        return self.britive.get(self.base_url, params=params)

    def get_vault_by_id(self, vault_id: str) -> dict:
        """
        Provide details of the given vault, from a vault id.

        :param vault_id: The ID  of the vault.
        :return: Details of the specified vault.
        """

        return self.britive.get(f'{self.base_url}/{vault_id}')

    def create(
        self,
        name: str,
        description: str = 'Default vault description',
        rotation_time: int = 30,
        encryption_algorithm: str = 'AES_256',
        default_notification_medium_id: str = '',
        users: list = [],
        tags: list = [],
        channels: list = [],
    ) -> dict:

        """
        Create a new vault.

        :param name: the name of the vault
        :param description: the description of the vault
        :param rotation_time: in hours, how often the vault should rotate keys
        :param encryption_algorithm : the encryption algorithm to use for the vault
        :param default_notification_medium_id : the default notification medium to use for the vault
        :param users: a list of user IDs to recieve notifications for the vault
        :param tags: a list of tags to recieve notifications for the vault
        :param channels : a list of channels to recieve notifications for the vault (only for slack)
        :return: Details of the newly created vault.
        """

        if default_notification_medium_id == '':
            for medium in self.britive.notification_mediums.list():
                if medium['name'] == 'Email':
                    default_notification_medium_id = medium['id']
        params = {
            'name': name,
            'description': description,
            'rotationTime': rotation_time,
            'encryptionAlgorithm': encryption_algorithm,
            'defaultNotificationMediumId': default_notification_medium_id,
            'recipients': {'userIds': users, 'tags': tags, 'channelIds': channels},
        }
        return self.britive.post(self.base_url, json=params)

    def delete(self, vault_id: str) -> None:
        """
        Deletes a vault.

        :param vault_id: the ID of the vault
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/{vault_id}')

    def update(self, vault_id: str, **kwargs) -> None:
        """
        Updates a vault.

        If not all kwargs a provided, the vault will update with the default values of the unprovided kwargs.

        :param vault_id: The ID of the vault.
        :param kwargs: Valid fields are...
            name - required
            description
            rotationTime - time in days between key rotations
            encryptionAlgorithm - the encryption algorithm to use for the vault
            defaultNotificationMediumId - the default notification medium to use for the vault
            recipients - a list of user IDs or tags to recieve notifications for the vault
        :return: None
        """

        return self.britive.patch(f'{self.base_url}/{vault_id}', json=kwargs)

    def rotate_keys(self) -> None:
        """
        Rotate vault keys.

        :return: None
        """

        return self.britive.post(
            f'{self.britive.base_url}/v1/secretmanager/keys/rotate'
        )


class PasswordPolicies:
    def __init__(self, britive) -> None:
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/v1/secretmanager/pwdpolicies'

    def get(self, password_policy_id: str) -> dict:
        """
        Provide details of the given password policy, from a password policy id.

        :param password_policy_id: The ID  of the password policy.
        :return: Details of the specified password policy.
        """

        return self.britive.get(f'{self.base_url}/{password_policy_id}')

    def list(self) -> list:
        """
        Provide a list of all password policies

        :return: List of all password policies
        """

        return self.britive.get(self.base_url)

    def create(
        self,
        name: str,
        description: str = 'Default description',
        password_type: str = 'alphanumeric',
        min_password_length: int = 8,
        has_upper_case_chars: bool = True,
        has_lower_case_chars: bool = True,
        has_numbers: bool = True,
        has_special_chars: bool = True,
        allowed_special_chars: str = '~`!@#$%^&*()-_+=[]{}|/;:"?/\\.><,\''
    ) -> dict:
        """
        Creates a new password policy.

        :param name: required, name of the password policy
        :param description: description of the password policy
        :param password_type: type of password to use for the policy
        :param min_password_length: minimum length of the password
        :param has_upper_case_chars: whether to require uppercase characters
        :param has_lower_case_chars: whether to require lowercase characters
        :param has_numbers: whether to require numbers
        :param has_special_chars: whether to require special characters
        :param allowed_special_chars: a string of special characters to allow in the password
        :return: Details of the newly created password policy.
        """

        params = {
            'name': name,
            'description': description,
            'passwordType': password_type,
            'minPasswordLength': min_password_length,
            'hasUpperCaseChars': has_upper_case_chars,
            'hasLowerCaseChars': has_lower_case_chars,
            'hasNumbers': has_numbers,
            'hasSpecialChars': has_special_chars,
            'allowedSpecialChars': allowed_special_chars,
        }
        return self.britive.post(self.base_url, json=params)

    def create_pin(self,  name: str,  description: str = 'Default description',  pin_length: int = 4) -> dict:
        """
        Creates a new pin password policy.

        :param name: required, name of the pin password policy
        :param description: description of the pin password policy
        :param pin_length: length of the pin to use for the policy
        :return: Details of the newly created pin password policy.
        """

        params = {
            'name': name,
            'description': description,
            'pinLength': pin_length,
            'passwordType': 'pin',
        }
        return self.britive.post(self.base_url, json=params)

    def update(self, password_policy_id: str, **kwargs) -> None:
        """
        Updates a password policy.

        :param password_policy_id: the ID of the password policy
        :param kwargs: Valid fields are...
            name: name of the password policy
            description: description of the password policy
            passwordType: type of password to use for the policy
            minPasswordLength: minimum length of the password
            hasUpperCaseChars: whether to require uppercase characters
            hasLowercaseChars: whether to require lowercase characters
            hasNumbers: whether to require numbers
            hasSpecialChars: whether to require special characters
            allowedSpecialChars: a string of special characters to allow in the password
            pinLength: the length of the pin to use for the policy (only for pins)
        :return: None
        """

        current = self.get(password_policy_id=password_policy_id)

        return self.britive.patch(f'{self.base_url}/{password_policy_id}', json={**current, **kwargs})

    def delete(self, password_policy_id: str) -> None:
        """
        Deletes a password policy.

        :param password_policy_id: the ID of the password policy
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/{password_policy_id}')

    def generate_password(self, password_policy_id: str) -> dict:
        """
        Generates a password for the given password policy.

        :param password_policy_id: the ID of the password policy
        :return: the generated password
        """

        params = {'action': 'generatePasswordOrPin'}
        return self.britive.get(f'{self.base_url}/{password_policy_id}', params=params)['passwordOrPin']

    def validate(self, password_policy_id: str, password: str) -> dict:
        """
        Validates a password for the given password policy.

        :param password_policy_id: the ID of the password policy
        :param password: the password to validate
        :return: whether the password is valid
        """

        data = {
            'id': password_policy_id,
            'passwordOrPin': password,
        }

        params = {
            'action': 'validatePasswordOrPin'
        }
        return self.britive.post(self.base_url, json=data, params=params)


class Folders:
    def __init__(self, britive) -> None:
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/v1/secretmanager/vault'

    def create(self, name: str, vault_id: str, path: str = '/') -> dict:
        """
        Creates a new folder in the vault.

        :param name: The name of the folder.
        :param vault_id: The ID of the vault.
        :param path: The path of the folder (include the  leading /).
        :return: Details of the newly created folder.
        """

        data = {'entityType': 'node', 'name': name}
        return self.britive.post(f'{self.base_url}/{vault_id}/secrets?path={path}', json=data)

    def delete(self, vault_id: str, path: str) -> None:
        """
        Deletes a folder from the vault.

        :param vault_id: ID of the vault to delete the folder from
        :param path: path of the folder, include the / at the beginning
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/{vault_id}/secrets?path={path}')


class Secrets:
    def __init__(self, britive) -> None:
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/v1/secretmanager/vault'

    def create(
        self,
        name: str,
        vault_id: str,
        path: str = '/',
        static_secret_template_id: str = '7a5f41d8-f7af-46a0-88f7-edf0403607ae',
        secret_mode: str = 'shared',
        secret_nature: str = 'static',
        value: dict = {'Note': 'This is the default note'},
        file: bytes = None
    ) -> dict:
        """
        Creates a new secret in the vault.

        For creating a secret with a file, you must read the file in with:
            with open(file_path, 'rb') as f:
                britive.secrets_manager.secrets.create(...., file=f)


        :param name: name of the secret
        :param vault_id: ID of the vault
        :param path: path of the secret, include the / the beginning
        :param static_secret_template_id: ID of the static secret template
        :param secret_mode: mode of the secret
        :param secret_nature: nature of the secret
        :param value: value of the secret (must be in the format of the static secret template)
        :param file: file to upload as the secret
        :return: Details of the newly created secret.
        """

        if not file:
            return self.britive.post(
                f'{self.base_url}/{vault_id}/secrets?path={path}',
                json={
                    'name': name,
                    'entityType': 'secret',
                    'staticSecretTemplateId': static_secret_template_id,
                    'secretMode': secret_mode,
                    'secretNature': secret_nature,
                    'value': value,
                }
            )
        else:
            secret_data = {
                'entityType': 'secret',
                'name': name,
                'staticSecretTemplateId': static_secret_template_id,
                'secretMode': secret_mode,
                'secretNature': secret_nature,
                'value': value,
            }
            return self.britive.post_upload(
                f'{self.base_url}/{vault_id}/secrets/file?path={path}',
                files={'file': file, 'secretData': (None, json.dumps(secret_data))},
            )

    def update(self, vault_id: str, path: str = '/', value: dict = {}) -> None:
        """
        Updates a secret's value

        :param vault_id: ID of the vault to update the secret in
        :param path: path of the secret, include the / at the beginning
        :param value: value of the secret
        :return: None
        """

        return self.britive.patch(
            f'{self.base_url}/{vault_id}/secrets?path={path}', json={'value': value}
        )

    def rename(self, vault_id: str, path: str = '/', new_name: str = '') -> None:
        """
        Update the name of a secret.

        :param vault_id: ID of the vault to update the secret in
        :param path: path of the secret, include the / at the beginning and the secret name
        :param new_name: new name of the secret
        :return: None
        """

        return self.britive.patch(
            f'{self.base_url}/{vault_id}/secrets?path={path}', json={'name': new_name}
        )

    def get(
        self,
        vault_id: str,
        path: str,
        secret_type: str = 'node',
        filter_type: str = None,
        recursive_secrets: bool = False,
        get_metadata: bool = True
    ) -> dict:
        """
        Gets a secret from the vault.

        :param vault_id: ID of the vault to get the secret from
        :param path: path of the secret, include the / at the beginning
        :param secret_type: type of the secret (node or secret)
        :param filter_type: filter to apply to the secret (NONE, ALL, SHARED, PRIVATE)
        :param recursive_secrets: whether or not to recursively get all secrets in the folder
        :param get_metadata: whether or not to get the metadata of the secret
        :return: Details of the secret.
        """

        params = {
            'type': secret_type,
            'filter': filter_type,
            'recursiveSecrets': (str(recursive_secrets)).lower(),
            'getMetadata': get_metadata,
        }
        return self.britive.get(
            f'{self.base_url}/{vault_id}/secrets?path={path}', params=params
        )

    def delete(self, vault_id: str, path: str) -> None:
        """
        Deletes a secret from the vault.

        :param vault_id: ID of the vault to delete the secret from
        :param path: path of the secret, include the / at the beginning
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/{vault_id}/secrets?path={path}')

    def access(self, vault_id: str, path: str, get_metadata: bool = False) -> dict:
        """
        Accesses a secret from the vault.

        :param vault_id: ID of the vault to get the secret from
        :param path: path of the secret, include the / at the beginning
        :param get_metadata: whether or not to get the metadata of the secret
        :return: Details of the secret.
        """

        params = {'getmetadata': get_metadata}
        return self.britive.get(
            f'{self.base_url}/{vault_id}/secrets?path={path}', params=params
        )


class Policies:
    def __init__(self, britive) -> None:
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/v1/policy-admin/policies'

    def list(self, path: str = '/', filter_str: str = None) -> dict:
        """
        Gets all policies in the vault.

        :param path: path of the policy, include the / at the beginning
        :param filter_str: filter to apply to the listing
        :return: Details of the policies.
        """

        params = {'resource': path, 'consumer': 'secretmanager'}
        if filter:
            params['filter'] = filter_str
        return self.britive.get(f'{self.base_url}', params=params)

    def delete(self, policy_id: str, path: str = '/') -> None:
        """
        Deletes a policy from the vault.

        :param policy_id: ID of the policy to delete
        :param path: path of the policy, include the / at the beginning
        :return: None
        """

        params = {'consumer': 'secretmanager', 'resource': path}
        return self.britive.delete(f'{self.base_url}/{policy_id}', params=params)

    def build(self, name: str, access_level: str = None, description: str = '', draft: bool = False,
              active: bool = True, read_only: bool = False, users: list = None, tags: list = None, tokens: list = None,
              service_identities: list = None, ips: list = None, from_time: str = None, to_time: str = None,
              date_schedule: dict = None, days_schedule: dict = None,
              approval_notification_medium: Union[str, list] = None, time_to_approve: int = 5,
              access_validity_time: int = 120, approver_users: list = None, approver_tags: list = None,
              access_type: str = 'Allow', identifier_type: str = 'name', condition_as_dict: bool = False) -> dict:
        """
        Build a policy document given the provided inputs.

        :param name: The name of the policy.
        :param access_level: The level of access. Valid values are SM_View, SM_Manage, SM_CRUD. Defaults to SM_View
            if not provided.
        :param description: An optional description of the policy.
        :param draft: Indicates if the policy is a draft. Defaults to `False`.
        :param active: Indicates if the policy is active. Defaults to `True`.
        :param read_only: Indicates if the policy is a read only. Defaults to `False`.
        :param users: Optional list of user names or ids to which this policy applies.
        :param tags: Optional list of tag names or ids to which this policy applies.
        :param tokens: Optional list of token names or ids to which this policy applies.
        :param service_identities: Optional list of service identity names or ids to which this policy applies.
        :param ips: Optional list of IP addresses for which this policy applies. Provide in CIDR notation
            or dotted decimal format for individual (/32) IP addresses.
        :param from_time: The start date/time of when the policy is in effect. If a date is provided
            (`YYYY-MM-DD HH:MM:SS`) this will represent the start date/time of 1 contiguous time range. If just a
            time is provided (`HH:MM:SS`) this will represent the daily recurring start time. If this parameter is
            provided then `to_time` must also be provided. This parameter is deprecated as of v2.19.0. The presence of
            `date_schedule` and/or `days_schedule` will override this field.
        :param to_time: The end date/time of when the policy is in effect. If a date is provided
            (`YYYY-MM-DD HH:MM:SS`) this will represent the end date/time of 1 contiguous time range. If just a
            time is provided (`HH:MM:SS`) this will represent the daily recurring end time. If this parameter is
            provided then `from_time` must also be provided. This parameter is deprecated as of v2.19.0. The presence of
            `date_schedule` and/or `days_schedule` will override this field.
        :param date_schedule: A dict in the format
            {
                'fromDate': '2022-10-29 10:30:00',
                'toDate': '2022-11-05 18:30:00',
                'timezone': 'UTC'
            }
            Timezone formats can be found in the TZ Identifier column of the following page.
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
        :param days_schedule: A dict in the format
            {
                'fromTime': '10:30:00',
                'toTime': '18:30:00',
                'timezone': 'UTC',
                'days': ['MONDAY','TUESDAY','WEDNESDAY','THURSDAY','FRIDAY','SATURDAY','SUNDAY']
            }
            Timezone formats can be found in the TZ Identifier column of the following page.
            https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
        :param approval_notification_medium: Optional notification medium name to which approval requests will be
            delivered. Can also specify a list of notification medium names. Specifying this parameter indicates the
            desire to enable approvals for this policy.
        :param time_to_approve: Optional number of minutes to wait for an approval before denying the action. Defaults
            to 5 minutes.
        :param access_validity_time: Optional number of minutes the access is valid after approval. Defaults to 120
            minutes.
        :param approver_users: Optional list of user names or ids who are to be considered approvers.
            If `approval_notification_medium` is set then either `approver_users` or `approver_tags` is required.
        :param approver_tags: Optional list of tag names who are considered approvers.
            If `approval_notification_medium` is set then either `approver_users` or `approver_tags` is required.
        :param access_type: The type of access this policy provides. Valid values are `Allow` and `Deny`. Defaults
            to `Allow`.
        :param identifier_type: Valid values are `id` or `name`. Defaults to `name`. Represents which type of
            identifiers are being provided to the other parameters. Either all identifiers must be names or all
            identifiers must be IDs.
        :param condition_as_dict: Prior to version 2.22.0 the only acceptable format for the condition block of
            a policy was as a stringifed json object. As of 2.22.0 the condition block can also be built as a raw
            python dictionary. This parameter will default to `False` to support backwards compatibility. Setting to
            `True` will result in the policy condition being returned/built as a python dictionary.
        :return: A dict which can be provided as a secret manager policy to `create` and `update`.
        """

        policy = self.britive.system.policies.build(
            name=name,
            description=description,
            draft=draft,
            active=active,
            read_only=read_only,
            users=users,
            tags=tags,
            tokens=tokens,
            service_identities=service_identities,
            ips=ips,
            from_time=from_time,
            to_time=to_time,
            date_schedule=date_schedule,
            days_schedule=days_schedule,
            approval_notification_medium=approval_notification_medium,
            time_to_approve=time_to_approve,
            access_validity_time=access_validity_time,
            approver_users=approver_users,
            approver_tags=approver_tags,
            access_type=access_type,
            identifier_type=identifier_type,
            condition_as_dict=condition_as_dict
        )

        policy.pop('permissions', None)
        policy.pop('roles', None)
        policy['accessLevel'] = access_level or 'SM_View'
        policy['consumer'] = 'secretmanager'
        return policy

    def create(self, policy: dict, path: str) -> dict:
        """
        Creates  a policy in the vault.

        :param policy: policy to create
        :param path: path of the policy, include the / at the beginning
        :return: Details of the policy.
        """

        policy['resource'] = path

        return self.britive.post(f'{self.base_url}?resource={path}&consumer=secretmanager', json=policy)


class StaticSecretTemplates:
    def __init__(self, britive) -> None:
        self.britive = britive
        self.base_url = (
            f'{self.britive.base_url}/v1/secretmanager/secret-templates/static'
        )

    def get(self, secret_template_id: str) -> dict:
        """
        Gets a secret template from the vault.

        :param secret_template_id: ID of the secret template to get
        :return: Details of the secret template.
        """

        return self.britive.get(f'{self.base_url}/{secret_template_id}')

    def list(self, filter_str: str = None) -> dict:
        """
        Lists all secret templates in the vault.

        :param filter_str: filter to apply to the listing
        :return: Details of the secret templates.
        """

        params = {'filter': filter_str}
        return self.britive.get(f'{self.base_url}', params=params)

    def delete(self, secret_template_id: str) -> None:
        """
        Deletes a secret template from the vault.

        :param secret_template_id: ID of the secret template to delete
        :return: None
        """

        return self.britive.delete(f'{self.base_url}/{secret_template_id}')

    def create(
        self,
        name: str,
        password_policy_id: str,
        description: str = '',
        rotation_interval: int = 30,
        parameters: list = None
    ) -> dict:
        """
        Creates a secret template

        :param name: name of the secret template
        :param password_policy_id: ID of the password policy to use
        :param description: description of the secret template
        :param rotation_interval: rotation interval of the secret template
        :param parameters: list of parameters to use in the secret template
        :return: Details of the secret template.
        """

        params = {
            'secretType': name,
            'passwordPolicyId': password_policy_id,
            'description': description,
            'rotationInterval': rotation_interval,
            'parameters': [parameters],
        }

        return self.britive.post(f'{self.base_url}', json=params)

    def update(self, static_secret_template_id: str, **kwargs) -> None:
        """
        Updates a secret template

        :param static_secret_template_id: ID of the secret template to update
        :param kwargs: key-value pairs to update the secret template with
                valid keys are:
                    name: name of the secret template
                    passwordPolicyId: ID of the password policy to use
                    description: description of the secret template
                    rotationInterval: rotation interval of the secret template
                    parameters: list of parameters to use in the secret template
        :return: None
        """

        current = self.get(secret_template_id=static_secret_template_id)

        return self.britive.patch(
            f'{self.base_url}/{static_secret_template_id}', json={**current, **kwargs}
        )


class Resources:
    def __init__(self, britive) -> None:
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/v1/secretmanager/resourceContainers'

    def get(self, path: str = '/') -> dict:
        """
        Gets a resource from the vault

        :param path: path of the resource, include the / at the beginning
        :return: Details of the resource.
        """

        params = {'path': path}
        return self.britive.get(f'{self.base_url}', params=params)



from datetime import datetime


class SettingsBanner:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/settings/banner'

    def get(self) -> dict:
        """
        Get the banner configuration.

        :returns: Details of the banner.
        """
        return self.britive.get(self.base_url)
    
    def set(self, display_banner: bool, message: str, message_type: str, start_datetime: datetime = None,
            end_datetime: datetime = None, time_zone: str = None) -> dict:
        """
        Sets the banner.

        :param display_banner: Whether to display the banner.
        :param message: The banner message.
        :param message_type: The banner type. This will drive the styling. Valid values are `INFO`,
            `WARNING`, and `CAUTION`.
        :param start_datetime: Optional start date and time for when the banner should be displayed.
            No timezone information should be encoded. If provided, must also provide `end_datetime` and `timezone`.
        :param end_datetime: Optional end date and time for when the banner should be displayed.
            No timezone information should be encoded. If provided, must also provide `start_datetime` and `timezone`.
        :param time_zone: Optional ISO timezone under which `start_datetime` and `end_datetime` should be interpreted.
            If provided, must also provide `start_datetime` and `end_datetime`.
        :returns: Details of the banner.
        """

        data = {
            'status': 'ON' if display_banner else 'OFF',
            'message': message,
            'messageType': message_type
        }

        schedule_fields = [start_datetime, end_datetime, time_zone]
        all_schedule_fields_present = all(v is not None for v in schedule_fields)
        no_schedule_fields_present = all(v is None for v in schedule_fields)

        if not all_schedule_fields_present and not no_schedule_fields_present:
            raise ValueError('if providing schedule information then start_datetime, '
                             'end_datetime, and time_zone are required')

        if all_schedule_fields_present:
            data['messageSchedule'] = {
                'endDate': end_datetime.isoformat(sep=' ', timespec='seconds'),
                'startDate': start_datetime.isoformat(sep=' ', timespec='seconds'),
                'timeZone': time_zone
            }

        return self.britive.post(self.base_url, json=data)



from .cache import *  # will also import some globals like `britive`


def test_list():
    secrets = britive.my_secrets.list()
    assert isinstance(secrets, list)
    assert len(secrets) > 0
    assert isinstance(secrets[0], dict)


def test_view_no_approval(cached_secret):
    data = britive.my_secrets.view(path=cached_secret['path'])
    assert isinstance(data, dict)



import os
import time
import requests
import json as native_json
import pkg_resources
import socket
from .helpers import methods as helper_methods
from .helpers import federation_providers as fp
from .users import Users
from .service_identity_tokens import ServiceIdentityTokens
from .service_identities import ServiceIdentities
from .exceptions import *
from .tags import Tags
from .applications import Applications
from .environments import Environments
from .environment_groups import EnvironmentGroups
from .scans import Scans
from .accounts import Accounts
from .permissions import Permissions
from .groups import Groups
from .identity_attributes import IdentityAttributes
from .profiles import Profiles
from .task_services import TaskServices
from .tasks import Tasks
from .security_policies import SecurityPolicies
from .saml import Saml
from .api_tokens import ApiTokens
from .audit_logs import AuditLogs
from .reports import Reports
from .identity_providers import IdentityProviders
from .my_access import MyAccess
from .notifications import Notifications
from .my_secrets import MySecrets
from .policies import Policies
from .secrets_manager import SecretsManager
from .notification_mediums import NotificationMediums
from .workload import Workload
from .system.system import System
from .settings.settings import Settings

BRITIVE_TENANT_ENV_NAME = 'BRITIVE_TENANT'
BRITIVE_TOKEN_ENV_NAME = 'BRITIVE_API_TOKEN'


class Britive:
    """
    Pure Python implementation for interacting with the Britive API.

    The end user should never need to use any other classes in this package. Everything is accessed via this `Britive`
    class.

    This package aims to wrap the Britive API for use in Python. For the most part it is a simple wrapper (sending
    potentially bad parameters to the API) but there are a couple of places where liberties were taken to enhance
    the developer/end user experience. Some APIs may also be combined into one Python method with a parameter if and
    where it makes more sense to present the API that way.

    Authentication is handled solely via API tokens. The token must be provided in one of two methods.

    - Passed directly into the class constructor.
    - Injected as an environment variable into the execution context where this package is being run. The
       environment variable name must be BRITIVE_API_TOKEN.

    As of v2.5.0 a `Bearer` token can be provided as well. A `Bearer` token is generated as part of an interactive
    login process and is temporary in nature. This change is to allow for an upcoming Python CLI application.

    All Britive API tokens are authenticated against a specific Britive tenant. The name of the tenant must be provided
    in one of two methods.

    - Passed directly into the class constructor.
    - Injected as an environment variable into the execution context where this package is being run. The
       environment variable name must be BRITIVE_TENANT.

    In order to obtain the tenant name, reference the Britive URL used to login to the UI. If the URL is
    https://example.britive-app.com then the tenant name will be `example`.

    No assumptions are made about the operating system or file system. Nothing is persisted to disk. The end user
    must persist responses to disk if and when that is required.
    """

    def __init__(self, tenant: str = None, token: str = None, query_features: bool = True,
                 token_federation_provider: str = None, token_federation_provider_duration_seconds: int = 900):
        """
        Instantiate an authenticated interface that can be used to communicate with the Britive API.

        :param tenant: The name of the Britive tenant. If the url you use to login to your Britive tenant is
            https://example.britive-app.com then your tenant name is `example` and is what you would provide here.
            If not provided then environment variable BRITIVE_TENANT will be used.
        :param token: The API token. If not provided then environment variable BRITIVE_API_TOKEN will be used.
        :param query_features: Indicates whether the SDK will query for features of the tenant (things like profile v1
            vs v2, secrets manager enabled,etc.). True by default but can be disabled as needed if the end user does not
            want to wait for that API call. Querying for features will help instruct the SDK as to what API calls are
            allowed to be used based on the features enabled, vs. attempting to make the API call and getting an error.
        :param token_federation_provider: The federation provider to use to source the token. Details of what can be
            provided can be found in the documentation for the Britive.source_federation_token_from method.
        :param token_federation_provider_duration_seconds: Only applicable for the AWS provider. Specify the number of
            seconds for which the generated token is valid. Defaults to 900 seconds (15 minutes).
        :raises: TenantMissingError, TokenMissingError
        """

        self.tenant = tenant or os.environ.get(BRITIVE_TENANT_ENV_NAME)

        if not self.tenant:
            raise TenantMissingError(
                'Tenant not explicitly provided and could not be sourced '
                f'from environment variable {BRITIVE_TENANT_ENV_NAME}'
            )

        if token_federation_provider:
            self.__token = self.source_federation_token_from(
                provider=token_federation_provider,
                tenant=self.tenant,
                duration_seconds=token_federation_provider_duration_seconds
            )
        else:
            self.__token = token or os.environ.get(BRITIVE_TOKEN_ENV_NAME)

        if not self.__token:
            raise TokenMissingError(
                'Token not explicitly provided and could not be sourced '
                f'from environment variable {BRITIVE_TOKEN_ENV_NAME}'
            )

        # clean up and apply logic to the passed in tenant (for backwards compatibility with no domain being required)
        self.tenant = self.parse_tenant(self.tenant)

        self.base_url = f'https://{self.tenant}/api'
        self.session = requests.Session()
        self.retry_max_times = 5
        self.retry_backoff_factor = 1
        self.retry_response_status = [429, 500, 502, 503, 504]

        # allow the disabling of TLS/SSL verification for testing in development (mostly local development)
        if os.getenv('BRITIVE_NO_VERIFY_SSL') and '.dev.' in self.tenant:
            # turn off ssl verification
            self.session.verify = False
            # wipe these due to this bug: https://github.com/psf/requests/issues/3829
            os.environ['CURL_CA_BUNDLE'] = ""
            os.environ['REQUESTS_CA_BUNDLE'] = ""
            # disable the warning message
            import urllib3
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        token_type = 'TOKEN' if len(self.__token) < 50 else 'Bearer'
        if len(self.__token.split('::')) > 1:
            token_type = 'WorkloadToken'

        try:
            version = pkg_resources.get_distribution('britive').version
        except Exception:
            version = 'unknown'

        self.session.headers.update({
            'Authorization': f'{token_type} {self.__token}',
            'Content-Type': 'application/json',
            'User-Agent': f'britive-python-sdk/{version} {requests.utils.default_user_agent()}'
        })

        self.feature_flags = self.features() if query_features else {}

        self.users = Users(self)
        self.service_identity_tokens = ServiceIdentityTokens(self)
        self.service_identities = ServiceIdentities(self)
        self.tags = Tags(self)
        self.applications = Applications(self)
        self.environments = Environments(self)
        self.environment_groups = EnvironmentGroups(self)
        self.scans = Scans(self)
        self.accounts = Accounts(self)
        self.permissions = Permissions(self)
        self.groups = Groups(self)
        self.identity_attributes = IdentityAttributes(self)
        self.profiles = Profiles(self)
        self.task_services = TaskServices(self)
        self.tasks = Tasks(self)
        self.security_policies = SecurityPolicies(self)
        self.saml = Saml(self)
        self.api_tokens = ApiTokens(self)
        self.audit_logs = AuditLogs(self)
        self.reports = Reports(self)
        self.identity_providers = IdentityProviders(self)
        self.my_access = MyAccess(self)
        self.notifications = Notifications(self)
        self.my_secrets = MySecrets(self)
        self.policies = Policies(self)
        self.secrets_manager = SecretsManager(self)
        self.notification_mediums = NotificationMediums(self)
        self.workload = Workload(self)
        self.system = System(self)
        self.settings = Settings(self)

    @staticmethod
    def source_federation_token_from(provider: str, tenant: str = None, duration_seconds: int = 900) -> str:
        """
        Returns a token from the specified federation provider.

        The caller must persist this token if required. New tokens can be generated on each invocation
        of this class as well.

        This method only works when running within the context of the specified provider.
        It is meant to abstract away the complexities of obtaining a federation token
        from common federation providers. Other provider federation tokens can still be
        sourced outside of this SDK and provided as input via the standard token presentation
        options.

        Six federation providers are currently supported by this method.

        * AWS IAM/STS, with optional profile specified - (aws)
        * Github Actions (github)
        * Bitbucket Pipelines (bitbucket)
        * Azure System Assigned Managed Identities (azuresmi)
        * Azure User Assigned Managed Identities (azureumi)
        * spacelift.io (spacelift)

        Any other OIDC federation provider can be used and tokens can be provided to this class for authentication
        to a Britive tenant. Details of how to construct these tokens can be found at https://docs.britive.com.

        :param provider: The name of the federation provider. Valid options are `aws`, `github`, `bitbucket`,
            `azuresmi`, `azureumi`, and `spacelift`.

            For the AWS provider it is possible to provide a profile via value `aws-profile`. If no profile is provided
            then the boto3 `Session.get_credentials()` method will be used to obtain AWS credentials, which follows
            the order provided here:
            https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html#configuring-credentials

            For the Github provider it is possible to provide an OIDC audience value via `github-<audience>`. If no
            audience is provided the default Github audience value will be used.

            For Azure User Assigned Managed Identities (azureumi) a client id is required. It must be
            provided in the form `azureumi-<client-id>`. From the Azure documentation...a user-assigned identity's
            client ID or, when using Pod Identity, the client ID of an Azure AD app registration. This argument
            is supported in all hosting environments.

            For both Azure Managed Identity options it is possible to provide an OIDC audience value via
            `azuresmi-<audience>` and `azureumi-<client-id>|<audience>`. If no audience is provided the default audience
             of `https://management.azure.com/` will be used.

        :param tenant: The name of the tenant. This field is optional but if not provided then the tenant will be
            sourced from environment variable BRITIVE_TENANT. Knowing the actual tenant is required for the AWS
            federation provider. This field can be ignored for non AWS federation providers.
        :param duration_seconds: Only applicable for the AWS provider. Specify the number of seconds for which the
            generated token is valid. Defaults to 900 seconds (15 minutes).
        :return: A federation token that can be used to authenticate to a Britive tenant.
        """

        helper = provider.split('-', maxsplit=1)
        provider = helper[0]

        if provider == 'aws':
            profile = helper_methods.safe_list_get(helper, 1, None)
            return fp.AwsFederationProvider(
                profile=profile,
                tenant=tenant,
                duration=duration_seconds
            ).get_token()

        if provider == 'github':
            audience = helper_methods.safe_list_get(helper, 1, None)
            return fp.GithubFederationProvider(audience=audience).get_token()

        if provider == 'bitbucket':
            return fp.BitbucketFederationProvider().get_token()

        if provider == 'azuresmi':
            audience = helper_methods.safe_list_get(helper, 1, None)
            return fp.AzureSystemAssignedManagedIdentityFederationProvider(audience=audience).get_token()

        if provider == 'azureumi':
            additional_attributes_str = helper_methods.safe_list_get(helper, 1, None)
            if not additional_attributes_str:
                raise ValueError('client id is required via azurumi-<client-id>')
            additional_attributes = additional_attributes_str.split('|')
            client_id = additional_attributes[0]
            audience = helper_methods.safe_list_get(additional_attributes, 1, None)
            return fp.AzureUserAssignedManagedIdentityFederationProvider(
                client_id=client_id,
                audience=audience
            ).get_token()

        if provider == 'spacelift':
            return fp.SpaceliftFederationProvider().get_token()

        raise InvalidFederationProvider(f'federation provider {provider} not supported')

    @staticmethod
    def parse_tenant(tenant: str) -> str:
        domain = tenant.replace('https://', '').replace('http://', '')   # remove scheme
        domain = domain.split('/')[0]  # remove any paths as they will not be needed
        try:
            domain_helper = domain.split(':')
            port = 443
            if len(domain_helper) > 1:
                port = domain_helper[1]
            domain_without_port = domain_helper[0]
            socket.getaddrinfo(host=domain_without_port, port=port)  # if success then a full domain was provided
            return domain
        except socket.gaierror:  # assume just the tenant name was provided (originally the only supported method)
            domain = f'{tenant}.britive-app.com'
            try:
                socket.getaddrinfo(host=domain, port=443)  # validate the hostname is real
                return domain  # and if so set the tenant accordingly
            except socket.gaierror:
                raise Exception(f'Invalid tenant provided: {tenant}. DNS resolution failed.')

    def features(self):
        features = {}
        for feature in self.get(f'{self.base_url}/features'):
            features[feature['name']] = feature['enabled']
        return features

    def banner(self) -> dict:
        return self.get(f'{self.base_url}/banner')

    def get(self, url, params=None):
        """Internal use only."""

        return self.__request('get', url, params=params)

    def post(self, url, params=None, data=None, json=None):
        """Internal use only."""

        return self.__request('post', url, params=params, data=data, json=json)

    def patch(self, url, params=None, data=None, json=None):
        """Internal use only."""

        return self.__request('patch', url, params=params, data=data, json=json)

    def put(self, url, params=None, data=None, json=None):
        """Internal use only."""

        return self.__request('put', url, params=params, data=data, json=json)

    def delete(self, url, params=None, data=None, json=None):
        """Internal use only."""

        return self.__request('delete', url, params=params, data=data, json=json)

    # note - this method could be iffy in the future if the app changes the way it handles
    # file uploads. As of 2022-01-26 it is working fine with the "Upload SAML Metadata" action
    # of an identity provider. This is the only use case currently.
    def patch_upload(self, url, file_content_as_str, content_type, filename):
        """Internal use only."""

        files = {
            filename: (f'{filename}.xml', file_content_as_str, content_type)
        }
        response = self.session.patch(url, files=files, headers={'Content-Type': None})
        try:
            return response.json()
        except native_json.decoder.JSONDecodeError:  # if we cannot decode json then the response isn't json
            return response.content.decode('utf-8')
    
    # note - this method is only used to upload a file when creating a secret
    def post_upload(self, url, params=None, files=None):
        """Internal use only."""
        response = self.session.post(url, params=params, files=files, headers={'Content-Type': None})
        try:
            return response.json()
        except native_json.decoder.JSONDecodeError:  # if we cannot decode json then the response isn't json
            return response.content.decode('utf-8')

    @staticmethod
    def __check_response_for_error(response):
        if response.status_code in allowed_exceptions.keys():
            try:
                content = native_json.loads(response.content.decode('utf-8'))
                message = f"{response.status_code} - " \
                          f"{content.get('errorCode') or 'E0000'} - " \
                          f"{content.get('message') or 'no message available'}"
                if content.get('details'):
                    message += f" - {content.get('details')}"
                raise allowed_exceptions[response.status_code](message)
            except native_json.decoder.JSONDecodeError as e:
                content = response.content.decode('utf-8')
                message = f"{response.status_code} - {content}"
                raise allowed_exceptions[response.status_code](message)

    @staticmethod
    def __response_has_no_content(response):
        # handle 204 No Content response
        if response.status_code == 204:
            return True

        # handle empty 200 response
        if response.status_code == 200 and len(response.content) == 0:
            return True

        return False

    @staticmethod
    def __pagination_type(headers, result):
        is_dict = isinstance(result, dict)
        has_next_page_header = 'next-page' in headers.keys()

        if is_dict and all(x in result.keys() for x in ['count', 'page', 'size', 'data']):
            return 'inline'
        if is_dict and has_next_page_header and all(x in result.keys() for x in ['data', 'reportId']):  # reports
            return 'report'
        if has_next_page_header:  # this interesting way of paginating is how audit_logs.query() does it
            return 'audit'
        if is_dict and all(x in result.keys() for x in ['result', 'pagination']):
            return 'secmgr'
        return 'none'

    @staticmethod
    def __tenant_is_under_maintenance(response):
        try:
            return response.status_code == 503 and response.json().get('errorCode') == 'MAINT0001'
        except:
            return False

    def __request_with_exponential_backoff_and_retry(self, method, url, params, data, json):
        num_retries = 0
        response = {}
        while num_retries <= self.retry_max_times:
            response = self.session.request(method, url, params=params, data=data, json=json)

            # handle the use case of a tenant being in maintenance mode
            # which means we should break out of this loop early and
            # not perform the backoff and retry logic
            if self.__tenant_is_under_maintenance(response=response):
                raise TenantUnderMaintenance(response.json().get('message'))

            # check for error - if so perform exponential backoff
            if response.status_code in self.retry_response_status:
                wait_time = (2 ** num_retries) * self.retry_backoff_factor
                time.sleep(wait_time)
                num_retries += 1
            else:  # we got a "good" response so break out of while loop
                break
        self.__check_response_for_error(response)  # handle an error response
        return response

    def __request(self, method, url, params=None, data=None, json=None):
        return_data = []
        num_iterations = 1
        pagination_type = None
        while True:  # infinite loop in case of pagination - we will break the loop when needed
            response = self.__request_with_exponential_backoff_and_retry(
                method=method,
                url=url,
                params=params,
                data=data,
                json=json
            )
            if self.__response_has_no_content(response):  # handle no content responses
                return None

            # handle secrets file download
            lowercase_headers = {h.lower(): v.lower() for h, v in response.headers.items()}
            content_disposition = lowercase_headers.get('content-disposition', '')
            if 'attachment' in content_disposition and 'downloadfile' in url:
                filename = response.headers.get('content-disposition').split('=')[1].replace('"', '').strip()
                return {'filename': filename, 'content_bytes': bytes(response.content)}

            # load the result as a dict
            try:
                result = response.json()
            except ValueError:  # includes simplejson.decoder.JSONDecodeError and native_json.decoder.JSONDecodeError
                return response.content.decode('utf-8')

            # check on the pagination and iterate if required - we only need to check on this after the first
            # request - checking it each time can screw up the logic when dealing with pagination coming from
            # the response headers as the header won't exist which will mean pagination_type will change to 'none'
            # which means we drop into the else block below and assign just the LAST page as the result, which
            # is obviously not what we want to be doing.
            if num_iterations == 1:
                pagination_type = self.__pagination_type(response.headers, result)

            if pagination_type == 'inline':
                return_data += result['data']
                count = result['count']
                page = result['page']
                size = result['size']
                if size * (page + 1) >= count:  # if we have reached the max number of records time to break the loop
                    break
                else:  # else loop again after incrementing the page number by 1
                    params['page'] = page + 1
            elif pagination_type == 'audit':
                return_data += result  # result is already a list
                if 'next-page' not in response.headers.keys():
                    break
                url = response.headers['next-page']
                params = {}  # the next-page header has all the URL parameters we need so unset them here
            elif pagination_type == 'report':
                return_data += result['data']
                if 'next-page' not in response.headers.keys():
                    break
                url = response.headers['next-page']
                params = {}  # the next-page header has all the URL parameters we need so unset them here
            elif pagination_type == 'secmgr':
                return_data += result['result']
                url = result['pagination'].get('next', '')
                if url == '':
                    break
            else:  # we are not dealing with pagination so just return the response as-is
                return_data = result
                break

            num_iterations += 1

        # finally return the response data
        return return_data

    def get_root_environment_group(self, application_id: str) -> str:
        """Internal use only."""

        app = self.applications.get(application_id=application_id)
        root_env_group = app.get('rootEnvironmentGroup') or {}
        for group in root_env_group.get('environmentGroups', []):
            if group['parentId'] == '':
                return group['id']
        raise RootEnvironmentGroupNotFound()




from .cache import *  # will also import some globals like `britive`


def test_get(cached_task_service):
    assert isinstance(cached_task_service, dict)
    assert cached_task_service['name'] == 'environmentScanner'
    assert cached_task_service['taskType'] == 'ApplicationScanner'


def test_enable(cached_task_service):
    response = britive.task_services.enable(task_service_id=cached_task_service['taskServiceId'])
    assert isinstance(response, dict)
    assert response['enabled']


def test_disable(cached_task_service):
    response = britive.task_services.disable(task_service_id=cached_task_service['taskServiceId'])
    assert isinstance(response, dict)
    assert not response['enabled']



class SystemConsumers:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/v1/policy-admin/consumers'

    def list(self) -> list:
        """
        List system level consumers.

        :returns: List of consumers.
        """

        return self.britive.get(self.base_url)['result']





from .cache import *  # will also import some globals like `britive`


def test_settings():
    settings = britive.saml.settings()
    assert isinstance(settings, dict)
    assert 'issuer' in settings.keys()
    assert 'id' in settings.keys()
    assert 'signInUrl' in settings.keys()
    assert 'signOutUrl' in settings.keys()
    assert 'x509CertExpirationDate' in settings.keys()
    settings = britive.saml.settings(as_list=True)
    assert isinstance(settings, list)
    assert len(settings) == 1


def test_metadata():
    metadata = britive.saml.metadata()
    assert isinstance(metadata, str)
    assert 'ds:X509Certificate' in metadata


def test_download():
    certificate = britive.saml.certificate()
    assert isinstance(certificate, str)
    assert '-----BEGIN CERTIFICATE-----' in certificate




from .cache import *  # will also import some globals like `britive`


def test_service_identity_tokens_create(cached_service_identity_token):
    assert isinstance(cached_service_identity_token, dict)
    assert cached_service_identity_token['tokenExpirationDays'] == 90
    assert 'token' in cached_service_identity_token.keys()


def test_service_identity_tokens_update(cached_service_identity_token_updated):
    assert isinstance(cached_service_identity_token_updated, dict)
    assert cached_service_identity_token_updated['tokenExpirationDays'] == 45
    assert 'token' in cached_service_identity_token_updated.keys()


def test_service_identity_tokens_get(cached_service_identity):
    token = britive.service_identity_tokens.get(cached_service_identity['userId'])
    assert isinstance(token, dict)
    assert token['tokenExpirationDays'] == 45
    assert 'tokenExpirationDays' in token.keys()





from .cache import *  # will also import some globals like `britive`


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_list(cached_application, cached_environment):
    groups = britive.accounts.list(
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id']
    )
    assert isinstance(groups, list)
    assert len(groups) > 0
    assert isinstance(groups[0], dict)


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_permissions(cached_application, cached_environment, cached_group):
    permissions = britive.groups.permissions(
        group_id=cached_group['appPermissionId'],
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id']
    )

    assert isinstance(permissions, list)


@pytest.mark.skipif(scan_skip, reason=scan_skip_message)
def test_accounts(cached_application, cached_environment, cached_group):
    groups = britive.groups.accounts(
        group_id=cached_group['appPermissionId'],
        application_id=cached_application['appContainerId'],
        environment_id=cached_environment['id']
    )

    assert isinstance(groups, list)




class SystemActions:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/v1/policy-admin/actions'

    def list(self, consumer: str = None) -> list:
        """
        List system level actions.

        :param consumer: The consumer for which the list should be filtered.
        :returns: List of actions.
        """

        params = {}
        if consumer:
            params['filter'] = f'consumer eq {consumer}'
        return self.britive.get(self.base_url, params=params)['result']





from .cache import *  # will also import some globals like `britive`


def test_list():
    reports = britive.reports.list()
    assert isinstance(reports, list)
    assert len(reports) >= 10  # at least 10 built-in reports exist so will alway be 10 for sure


def test_run_json():
    for report in britive.reports.list():
        if report['name'] == 'Profile Last Access':
            report_id = report['reportId']

    report = britive.reports.run(report_id=report_id, csv=False)
    assert isinstance(report, list)
    assert len(report) > 0  # at least 1 row in the report
    assert(len(report)) % 100 != 0  # v2.8.1 - adding check due to pagination bug not including the last page


def test_run_csv():
    for report in britive.reports.list():
        if report['name'] == 'Profile Last Access':
            report_id = report['reportId']

    report = britive.reports.run(report_id=report_id, csv=True)
    assert isinstance(report, str)
    assert 'application,environment,applicationStatus' in report

